package com.linecorp.line.camera.controller.function.record

import android.app.Activity
import android.util.Log
import com.linecorp.line.camera.controller.function.record.action.GifRecordAction
import com.linecorp.line.camera.controller.function.record.action.MockRecordAction
import com.linecorp.line.camera.controller.function.record.action.PhotoVideoRecordAction
import com.linecorp.line.camera.controller.function.record.action.RecordAction
import com.linecorp.line.camera.core.ILineCamera
import com.linecorp.line.camera.model.CameraConstExt
import com.linecorp.line.media.BuildConfig
import com.linecorp.yuki.camera.effect.android.model.FaceStickerModel
import com.linecorp.yuki.camera.effect.android.model.ScreenFilterModel
import com.linecorp.yuki.camera.effect.android.util.CameraConst

class RecordActionController {

    companion object {
        private val TAG = "RecordActionController"
    }

    constructor(activity: Activity) {
        this.activity = activity
        gaManager = GaManager()
        recordAction = MockRecordAction()

        recordSoundManager = RecordSoundManager()
        recordSoundManager.initPlayer(activity)
    }

    private lateinit var lineCamera: ILineCamera
    private var activity: Activity
    private lateinit var photoVideoRecordAction: PhotoVideoRecordAction
    private lateinit var gifRecordAction: GifRecordAction
    private var recordAction: RecordAction
    private val recordSoundManager: RecordSoundManager
    private val gaManager: GaManager

    private var isCameraReady = false
    private val a11YStrRes: Int = 0

    val needLongPressed: Boolean
        get() = recordAction.getNeedLongPressed()

    fun processAfterOnClick() {
        if (recordAction.cancelCountdown()) {
            return
        }
        recordAction.takeShortRecord()
    }

    fun processAfterOnLongClick() {
        if (recordAction.cancelCountdown()) {
            return
        }
        recordAction.takeLongRecord()
    }

    fun onCameraReady() {
        isCameraReady = true
        recordAction.setIgnoreKeyEvent(false)
    }

    fun onChangedCameraMode() {
        recordAction = getRecordAction(lineCamera)
        recordAction.setIgnoreKeyEvent(!isCameraReady)
    }

    private fun getRecordAction(lineCamera: ILineCamera): RecordAction {
        return when (lineCamera.cameraMode) {
            CameraConstExt.CameraMode.GIF -> {
                gifRecordAction = GifRecordAction()
                gifRecordAction.init(activity)
                gifRecordAction.setSoundManager(recordSoundManager)
                gifRecordAction.setGaManager(gaManager)
                gifRecordAction.setLineCamera(lineCamera)
                gifRecordAction
            }

            CameraConstExt.CameraMode.PHOTO_OR_VIDEO -> {
                photoVideoRecordAction = PhotoVideoRecordAction()
                photoVideoRecordAction.init(activity)
                photoVideoRecordAction.setSoundManager(recordSoundManager)
                photoVideoRecordAction.setGaManager(gaManager)
                photoVideoRecordAction.setLineCamera(lineCamera)
                photoVideoRecordAction
            }
        }
    }

    fun onChangedGifCameraStatus(gifState: CameraConstExt.RecordGifState) {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "onChangedGifCameraStatus: " + gifState)
        }
        recordAction.onChangedGifCameraStatus(gifState)
    }

    fun onCancelTimer(): Boolean {
        return recordAction.cancelCountdown()
    }

    fun setIgnoreKeyEvent(ignoreKeyEvent: Boolean) {
        recordAction.setIgnoreKeyEvent(ignoreKeyEvent)
    }

    fun onKeyDown(): Boolean {
        if (recordAction.cancelCountdown()) {
            return true
        }
        return recordAction.onKeyDown()
    }

    fun onDestroy() {
        recordSoundManager.release()
    }

    fun onBackPressed(): Boolean {
        if (recordAction.isNeedToCancelRecording()) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onBackPressed() returned: (lineCamera.isVideoRecording() || lineCamera.getRecordState() == CameraConst.RecordState.PAUSE) or (Stop Gif Recording)")
            }
            recordAction.cancelRecording()
            return true
        } else if (recordAction.cancelCountdown()) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onBackPressed() returned: lineCamera.cancelCountdownInSeconds()")
            }
            return true
        }
        return false
    }

    fun setLineCamera(lineCamera: ILineCamera) {
        this.lineCamera = lineCamera
    }

    fun setScreenFilter(screenFilterModel: ScreenFilterModel?) {
        gaManager.setScreenFilter(screenFilterModel)
        gaManager.setScreenFilterModelStr(activity.resources.getString(screenFilterModel!!.filterName))
    }

    fun setFaceSticker(faceStickerModel: FaceStickerModel?) {
        gaManager.setFaceSticker(faceStickerModel)
    }

    fun setDistortionDensity(distortionDensity: Int) {
        gaManager.setDistortionDensity(distortionDensity)
    }

    fun setGaType(gaType: String) {
        gaManager.setGaType(gaType)
    }

    fun setFacing(facing: CameraConst.Facing) {
        gaManager.setFacing(facing)
    }

    fun stopRecording() {
        recordAction.stopRecording()
    }

    fun getA11YStrRes(): Int {
        return recordAction.getA11yStrRes()
    }
}
