package com.linecorp.line.camera.controller.function.record.action

import android.app.Activity
import android.os.Handler
import android.util.Log
import com.linecorp.line.camera.LineMixCamera
import com.linecorp.line.camera.controller.function.record.GaManager
import com.linecorp.line.camera.controller.function.record.RecordSoundManager
import com.linecorp.line.camera.controller.function.record.RecordSoundManager.Companion.PLAY_SOUND
import com.linecorp.line.camera.controller.function.record.RecordSoundManager.Companion.PLAY_STOP
import com.linecorp.line.camera.core.ILineCamera
import com.linecorp.line.camera.model.CameraConstExt
import com.linecorp.line.media.BuildConfig
import com.linecorp.line.media.R
import com.linecorp.line.media.analytics.MediaGAEvents
import com.linecorp.yuki.camera.effect.android.OnCompleteTimerForVideoRecordingListener
import com.linecorp.yuki.camera.effect.android.OnShutterToCaptureStillFrameListener
import com.linecorp.yuki.camera.effect.android.util.CameraConst
import jp.naver.line.androie.common.PermissionRequestActivity
import jp.naver.line.androie.common.access.LineAccessForCommonHelper
import jp.naver.line.androie.db.generalkv.dao.GeneralKey
import jp.naver.line.androie.util.PermissionUtils

open class PhotoVideoRecordAction : RecordAction {
    companion object {
        private const val TAG = "PhotoVideoRecordAction"
    }

    private lateinit var lineCamera: ILineCamera
    private val handler: Handler = Handler()
    private var activity: Activity? = null
    private lateinit var recordSoundManager: RecordSoundManager
    private lateinit var gaManager: GaManager

    private var ignoreKeyEvent = true

    override fun init(activity: Activity?) {
        this.activity = activity
    }

    override fun stopRecording() = cancelRecording()

    override fun cancelRecording() = lineCamera.cancelVideoRecording()

    override fun takeShortRecord() {
        val state = lineCamera.recordState
        when (lineCamera.shotMode) {
            CameraConst.ShotMode.ONLY_PHOTO -> processTakingPicture(lineCamera)
            CameraConst.ShotMode.PHOTO_VIDEO -> when (state) {
                CameraConst.RecordState.READY -> processTakingPicture(lineCamera)
                else -> processVideoRecording(lineCamera)
            }
            CameraConst.ShotMode.ONLY_VIDEO -> processVideoRecording(lineCamera)
        }
    }

    override fun takeLongRecord() {
        if (lineCamera.shotMode == CameraConst.ShotMode.PHOTO_VIDEO) {
            processVideoRecording(lineCamera)
        }
    }

    override fun onChangedGifCameraStatus(recordGifState: CameraConstExt.RecordGifState) = Unit

    override fun onKeyDown(): Boolean =
            if (lineCamera.shotMode != CameraConst.ShotMode.ONLY_VIDEO) {
                processTakingPicture(lineCamera)
            } else false

    override fun getNeedLongPressed(): Boolean =
            lineCamera.shotMode == CameraConst.ShotMode.PHOTO_VIDEO

    override fun setLineCamera(lineCamera: ILineCamera) {
        this.lineCamera = lineCamera
        if (lineCamera.shotMode == CameraConst.ShotMode.ONLY_VIDEO) {
            ignoreKeyEvent = true
        }
    }

    override fun setIgnoreKeyEvent(ignoreKeyEvent: Boolean) {
        this.ignoreKeyEvent = ignoreKeyEvent
    }

    override fun setSoundManager(recordSoundManager: RecordSoundManager) {
        this.recordSoundManager = recordSoundManager
    }

    override fun setGaManager(gaManager: GaManager) {
        this.gaManager = gaManager
    }

    override fun isNeedToCancelRecording(): Boolean =
            lineCamera.isVideoRecording || lineCamera.recordState == CameraConst.RecordState.PAUSE

    override fun getA11yStrRes(): Int = when (lineCamera.recordState) {
        CameraConst.RecordState.START, CameraConst.RecordState.RESUME, CameraConst.RecordState.PAUSE, CameraConst.RecordState.STOP -> R.string.access_camera_shoot
        CameraConst.RecordState.READY -> R.string.access_video_record_done
        else -> R.string.access_camera_shoot
    }

    private fun processTakingPicture(lineCamera: ILineCamera): Boolean {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "processTakingPicture() called with: lineCamera = [$lineCamera]")
        }
        if (ignoreKeyEvent) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "processTakingPicture() returned:[ignoreKeyEvent == true] " + false)
            }
            return false
        }
        lineCamera.takePicture(OnShutterListener())
        ignoreKeyEvent = true
        gaManager.sendGaPhotoEvent()
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "processTakingPicture() returned: " + true)
        }
        return true
    }

    private fun processVideoRecording(lineCamera: ILineCamera) {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "processVideoRecording() called with: lineCamera = [$lineCamera]")
        }
        if (checkPermission()) {
            return
        }
        if (cancelCountdown()) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "processVideoRecording() returned: [cancelCountdown]")
            }
            return
        }

        when (lineCamera.recordState) {
            CameraConst.RecordState.START, CameraConst.RecordState.RESUME, CameraConst.RecordState.PAUSE, CameraConst.RecordState.STOP -> {
                lineCamera.stopVideoRecording()
                recordSoundManager.play(PLAY_STOP)

                LineAccessForCommonHelper.getLineAccess().sendGaEvent(
                        MediaGAEvents.MEDIA_CAMERA_RECORDING_DONE.category,
                        MediaGAEvents.MEDIA_CAMERA_RECORDING_DONE.action,
                        MediaGAEvents.MEDIA_CAMERA_RECORDING_DONE.label)
            }

            CameraConst.RecordState.READY -> {
                val listener = OnCompleteTimerForVideoRecordingListener {
                    val streamId = recordSoundManager.play(PLAY_SOUND)
                    handler.postDelayed({
                        if (streamId != 0) {
                            recordSoundManager.stop(streamId)
                            lineCamera.startVideoEncoding()
                        }
                    }, RecordSoundManager.PLAY_SOUND_DURATION_MILLISECONDS)
                }
                lineCamera.startVideoRecording(listener)

                gaManager.sendGaVideoEvent()
            }
            else ->
                return
        }
    }

    override fun cancelCountdown(): Boolean {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "cancelCountdown() called with: lineCamera = [$lineCamera]")
        }
        if (lineCamera.cancelCountdownInSeconds()) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "processTakingPicture() returned:[lineCamera.cancelCountdownInSeconds()] " + true)
            }
            setIgnoreKeyEvent(false)
            return true
        }
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "cancelCountdown() returned: " + false)
        }
        return false
    }

    private fun checkPermission(): Boolean {
        if (activity == null) {
            return false
        }
        if (!PermissionUtils.isPermissionGranted(activity as Activity, *LineMixCamera.PERMISSION_CUSTOM_CAMCORDER)) {
            LineAccessForCommonHelper.getLineAccess().putGeneralKeyValue(GeneralKey.MEDIA_CAMERA_CAMCORDER_PERMISSION_NOTICE, true)
            val intent = PermissionRequestActivity.createIntentForActivityResult(activity as Activity, *LineMixCamera.PERMISSION_CUSTOM_CAMCORDER)
            activity?.startActivity(intent)
            return true
        }
        return false
    }

    private inner class OnShutterListener : OnShutterToCaptureStillFrameListener {
        override fun onShutterToCaptureStillFrame() {
            recordSoundManager.play(RecordSoundManager.PLAY_SHUTTER)
        }
    }
}