package com.linecorp.line.camera.core.gif

import android.graphics.RectF
import android.opengl.GLES20
import android.util.Log
import com.linecorp.line.media.BuildConfig
import com.linecorp.opengl.TextureFrameBuffer
import com.linecorp.opengl.transform.ScaleType
import com.linecorp.opengl.util.FrameBufferSwapper
import com.linecorp.yuki.camera.effect.android.YukiCameraEffectService
import java.nio.ByteBuffer

class GifFrameExtractor(private var onChangedGifFrameExtractListener: OnChangedGifFrameExtractListener, private var yukiCameraEffectService: YukiCameraEffectService, var limitDuration: Int = 0) {
    companion object {
        private const val TAG: String = "GifFrameExtractor"
        private const val FRAME_SKIP_DURATION_MS = 90
    }

    var imageRotationDegree = 0

    var startTime: Long = 0
    var capturedFrameCnt = 0
    var captureTimeMillis: Long = 0

    private val gifFileSaver: GifFrameFileSaver = GifFrameFileSaver()
    private var onChangedGifFileSaveListener: OnChangedGifFileSaveListener

    private var isCanceledExtractFrame = false
    private var isStopExtractFrame = false

    fun startFrameExtract(imageRotationDegree: Int) {
        if (limitDuration == 0) {
            if (BuildConfig.DEBUG) {
                Log.e(TAG, "limitDuration can not be zero: ")
            }
            return
        }
        this.imageRotationDegree = imageRotationDegree

        // initialise the value
        isCanceledExtractFrame = false
        isStopExtractFrame = false
        capturedFrameCnt = 0
        captureTimeMillis = 0

        // start extract and save frames
        gifFileSaver.startBgFileSaver()
        gifFileSaver.registerOnChangedGifFileSaveListener(onChangedGifFileSaveListener)
        yukiCameraEffectService.setCameraEffectListener(YukiCameraEffectListener(limitDuration, onChangedGifFrameExtractListener))
    }

    fun cancelFrameExtract() {
        if (limitDuration == 0) {
            if (BuildConfig.DEBUG) {
                Log.e(TAG, "limitDuration can not be zero: ")
            }
            return
        }
        isCanceledExtractFrame = true
        gifFileSaver.requestCanceledSaveFiles()
        release()
    }

    fun stopFrameExtract() {
        if (limitDuration == 0) {
            if (BuildConfig.DEBUG) {
                Log.e(TAG, "limitDuration can not be zero: ")
            }
            return
        }
        isStopExtractFrame = true
    }

    private fun release() = yukiCameraEffectService.setCameraEffectListener(null)

    private inner class YukiCameraEffectListener() : YukiCameraEffectService.Listener {
        internal var limitDuration: Int = 0
        internal var totalRenderMs: Long = 0
        internal lateinit var onChangedGifFrameExtractListener: OnChangedGifFrameExtractListener

        constructor(limitDuration: Int, onChangedGifFrameExtractListener: OnChangedGifFrameExtractListener) : this() {
            startTime = System.currentTimeMillis()
            this.limitDuration = limitDuration
            this.onChangedGifFrameExtractListener = onChangedGifFrameExtractListener
            onChangedGifFrameExtractListener.onFrameExtractStarted()
        }

        override fun onFirstFrame() {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onFirstFrame: ")
            }
        }

        override fun onBeforeFrameRender() {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onBeforeFrameRender: ")
            }
        }

        override fun onFrameRender(frameBuffer: TextureFrameBuffer) {
            if (!gifFileSaver.isReady()) {
                if (BuildConfig.DEBUG) {
                    Log.d(TAG, "onFrameRender: gifFileSaver is not ready")
                }
                return
            }

            if (!frameBuffer.isAvailable) {
                if (BuildConfig.DEBUG) {
                    Log.d(TAG, "onFrameRender: frameBuffer is not alive")
                }
                return
            }

            val currentTimeMillis = System.currentTimeMillis()
            val needRetryForLongRecording = capturedFrameCnt < 2
            if (isStopExtractFrame) {
                if (BuildConfig.DEBUG) {
                    Log.d(TAG, "onFrameRender: stopExtractFrame : $isStopExtractFrame, but needMoreFrame to make gif $needRetryForLongRecording")
                }
                if (!needRetryForLongRecording) {
                    requestFinishSaveFile()
                    return
                }
            }

            if (isCanceledExtractFrame) {
                if (BuildConfig.DEBUG) {
                    Log.d(TAG, "onFrameRender: canceledExtractFrame : $isCanceledExtractFrame")
                }
                return
            }

            if (captureTimeMillis - startTime > limitDuration) {
                if (BuildConfig.DEBUG) {
                    Log.d(TAG, "onFrameRender: reached limit duration : ")
                }
                requestFinishSaveFile()
                return
            }

            if (currentTimeMillis < captureTimeMillis + FRAME_SKIP_DURATION_MS && !needRetryForLongRecording) {
                if (BuildConfig.DEBUG) {
                    Log.d(TAG, "onFrameRender() skip, the current duration is in : $FRAME_SKIP_DURATION_MS")
                }
                return
            }

            var w = frameBuffer.width
            var h = frameBuffer.height
            val numOfBytes = w * h * 4 // RGBA

            val byteBuffer = ByteBuffer.allocateDirect(numOfBytes)

            if (imageRotationDegree == 90 || imageRotationDegree == 270) {
                val temp = w
                w = h
                h = temp
            }

            // Swap frame for capturing with rotation.
            val captureFrameBuffer = TextureFrameBuffer(w, h)
            val captureFrameBufferSwapper = FrameBufferSwapper(frameBuffer.texture, captureFrameBuffer, null as RectF?, null as ScaleType?, imageRotationDegree, true, false)
            captureFrameBufferSwapper.swap()

            GLES20.glReadPixels(0, 0, w, h, GLES20.GL_RGBA, GLES20.GL_UNSIGNED_BYTE, byteBuffer)
            byteBuffer.limit(numOfBytes)

            requestSaveFile(w, h, byteBuffer)
            capturedFrameCnt++

            // For Logging frame counts and capturing speed
            captureTimeMillis = System.currentTimeMillis()
            val eachMs = captureTimeMillis - currentTimeMillis
            totalRenderMs += eachMs

            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onFrameRender() returned: eachMs : " + (captureTimeMillis - currentTimeMillis) + ", totalMs:" + (captureTimeMillis - startTime) + ", totalCnt:" + (capturedFrameCnt - 1))
            }

            captureFrameBuffer.release()
            captureFrameBufferSwapper.release()
        }

        override fun onSourceDraw(i: Int, i1: Int, i2: Int): Boolean {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onSourceDraw: ")
            }
            return false
        }

        override fun onMixDraw(i: Int, i1: Int, i2: Int): Boolean {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onMixDraw: ")
            }
            return false
        }

        override fun onFinalDraw(i: Int, i1: Int, i2: Int) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onFinalDraw: ")
            }
        }

        private fun requestFinishSaveFile() {
            if (capturedFrameCnt == 0) {
                onChangedGifFileSaveListener.onError(GifCameraErrorCode.ERROR_CAPTURED_FRAME_ZERO)
                return
            }
            gifFileSaver.requestSaveFinalFile((captureTimeMillis - startTime), capturedFrameCnt)
            onChangedGifFrameExtractListener.onCompleteExtractFrame()
            release()
        }

        private fun requestSaveFile(width: Int, height: Int, byteBuffer: ByteBuffer) =
                gifFileSaver.requestSaveFile(width, height, byteBuffer)
    }

    private inner class GifFileSaveListener : OnChangedGifFileSaveListener {
        override fun onCanceledSaveFile() =
                onChangedGifFrameExtractListener.onCanceledExtractFrame()

        override fun onError(errorCode: Int) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onError() : " + errorCode)
            }
            cancelFrameExtract()
            onChangedGifFrameExtractListener.onError(errorCode)
        }

        override fun onCompleteSaveFile(dstFileDir: String, fileCnt: Int) {
            if (fileCnt == 0) {
                onChangedGifFileSaveListener.onError(GifCameraErrorCode.ERROR_CAPTURED_FRAME_ZERO)
                return
            }
            onChangedGifFrameExtractListener.onCompleteSaveFile(dstFileDir, ((captureTimeMillis - startTime) / fileCnt))
        }
    }

    init {
        this.onChangedGifFileSaveListener = GifFileSaveListener()
    }
}
