package com.linecorp.line.camera.core.gif

import android.graphics.Bitmap
import android.os.Handler
import android.os.HandlerThread
import android.os.Looper
import android.os.Message
import android.text.TextUtils
import android.util.Log
import com.linecorp.line.camera.core.gif.GifCameraErrorCode.Companion.ERROR_FILE_EXCEPTION
import com.linecorp.line.common.LineMediaFileManager
import com.linecorp.line.media.BuildConfig
import java.io.File
import java.io.FileNotFoundException
import java.io.FileOutputStream
import java.nio.ByteBuffer

class GifFrameFileSaver {
    companion object {
        private const val TAG: String = "GifFrameFileSaver"
        private const val MSG_MAKE_FILE = 0
        private const val MSG_FINAL_FILE = 1
    }

    private var handlerThread: HandlerThread? = FileSaveHandlerThread("myHandleThread")
    private var onChangedGifFileSaveListener: OnChangedGifFileSaveListener? = null
    private var handler: Handler? = null
    var gifTempFolderPath: String = ""

    fun startBgFileSaver() {
        gifTempFolderPath = LineMediaFileManager.getGifCameraCacheSubDirPath(System.currentTimeMillis().toString()) ?: ""
        if (TextUtils.isEmpty(gifTempFolderPath)) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "GifFrameFileSaver gifTempFolderPath is null")
            }
            return
        }

        if (handlerThread == null) {
            handlerThread = FileSaveHandlerThread("myHandleThread")
        }

        handlerThread?.start()
    }

    private fun createGifCacheFile(byteBuffer: ByteBuffer, w: Int, h: Int) {
        try {
            // copy pixels
            val bitmap = Bitmap.createBitmap(w, h, Bitmap.Config.ARGB_8888)
            bitmap.copyPixelsFromBuffer(byteBuffer)

            val maxPixel = 450

            // scaled with the aspect ratio
            var width: Int
            var height: Int

            if (w > h) {
                width = maxPixel
                height = (maxPixel * getAspectRatio(w, h)).toInt()
            } else {
                width = (maxPixel * getAspectRatio(h, w)).toInt()
                height = maxPixel
            }

            val newBitmap = Bitmap.createScaledBitmap(bitmap, width, height, true)

            // Write Bitmap to File
            var fos: FileOutputStream? = null
            val dstFilePath = gifTempFolderPath + File.separator + System.currentTimeMillis() + ".jpg"

            if (BuildConfig.DEBUG) {
                Log.d(TAG, "createGifCacheFile: $dstFilePath\nwidth : $width, height : $height")
            }

            try {
                // compress to JPEG
                fos = FileOutputStream(dstFilePath)
                newBitmap.compress(Bitmap.CompressFormat.JPEG, 50, fos)
                bitmap.recycle()
                newBitmap.recycle()
            } finally {
                if (fos != null) {
                    fos.close()
                }
            }
            val file = File(dstFilePath)
            if (!file.exists()) {
                FileNotFoundException("dstFilePath is not exist. Somethings are wrong when take picture.")
            }
        } catch (e: Throwable) {
            onChangedGifFileSaveListener?.onError(ERROR_FILE_EXCEPTION)
            if (BuildConfig.DEBUG) {
                Log.e(TAG, e.message, e)
            }
        }
    }

    private fun getAspectRatio(x: Int, y: Int): Float {
        return y / x.toFloat()
    }

    private inner class FileSaveHandlerThread(name: String) : HandlerThread(name) {
        override fun onLooperPrepared() {
            super.onLooperPrepared()
            handler = FileSaveMsgHandler(looper)
        }
    }

    private inner class FileSaveMsgHandler(looper: Looper) : Handler(looper) {

        internal var totalTime: Long = 0
        internal var currentTime: Long = 0
        internal var cnt: Int = 0

        override fun handleMessage(msg: Message) {
            super.handleMessage(msg)

            val prevTime = System.currentTimeMillis()

            // create a file to JPEG
            when (msg.what) {
                MSG_MAKE_FILE -> createGifCacheFile(msg.obj as ByteBuffer, msg.arg1, msg.arg2)
                MSG_FINAL_FILE -> {
                    onChangedGifFileSaveListener?.onCompleteSaveFile(gifTempFolderPath, cnt)
                    handlerThread?.quit()
                    handlerThread = null
                }
            }

            // logging the count and time of saving file
            currentTime = System.currentTimeMillis()
            val eachTime = currentTime - prevTime
            totalTime += eachTime

            cnt++
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "handleMessage: eachMs : $eachTime, totalMs : $totalTime, cnt : $cnt")
            }
        }
    }

    fun isReady(): Boolean {
        if (handler == null) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "handler == null")
            }
            return false
        }

        if (handlerThread == null || handlerThread?.isAlive == false) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, " or handlerThread == null, handlerThread is not alive")
            }
            return false
        }
        return true
    }

    fun requestSaveFinalFile(totalRenderMs: Long, capturedFrameCnt: Int) {
        val message = Message.obtain()
        message.what = MSG_FINAL_FILE
        message.arg1 = totalRenderMs.toInt() / capturedFrameCnt
        handler?.sendMessage(message)
    }

    fun requestSaveFile(width: Int, height: Int, byteBuffer: ByteBuffer) {
        val message = Message.obtain()
        message.what = MSG_MAKE_FILE
        message.arg1 = width
        message.arg2 = height
        message.obj = byteBuffer

        handler?.sendMessage(message)
    }

    fun requestCanceledSaveFiles() {
        handler?.removeMessages(MSG_MAKE_FILE)
        handler?.removeMessages(MSG_FINAL_FILE)
        handlerThread?.quit()
        handlerThread = null
        onChangedGifFileSaveListener?.onCanceledSaveFile()
    }

    fun registerOnChangedGifFileSaveListener(onChangedGifFileSaveListener: OnChangedGifFileSaveListener) {
        this.onChangedGifFileSaveListener = onChangedGifFileSaveListener
    }
}