package com.linecorp.line.camera.mode.gif

import android.util.Log
import com.linecorp.line.camera.core.gif.GifFrameExtractor
import com.linecorp.line.camera.core.gif.OnChangedGifCameraStatusListener
import com.linecorp.line.camera.core.gif.OnChangedGifFrameExtractListener
import com.linecorp.line.camera.core.gif.timer.GifCameraTimer
import com.linecorp.line.camera.core.gif.timer.OnGifCameraTimerCountDownListener
import com.linecorp.line.camera.model.CameraConstExt
import com.linecorp.line.camera.model.CameraConstExt.RecordGifState
import com.linecorp.line.media.BuildConfig
import com.linecorp.yuki.camera.effect.android.YukiCameraEffectService
import com.linecorp.yuki.camera.effect.android.util.CameraConst

class GifCameraManager(yukiCameraEffectService: YukiCameraEffectService) {
    companion object {
        private const val TAG: String = "CameraModeManager"
        const val DEFAULT_RECORD_GIF_DURATION_MS = 3000
        const val LONG_RECORD_GIF_DURATION_MS = 8000
    }

    var gifFrameExtractor: GifFrameExtractor? = null
    private var onChangedGifFrameExtractListener: OnChangedGifFrameExtractListener

    var gifRecordStatus: CameraConstExt.RecordGifState = RecordGifState.READY
    lateinit var onChangedGifCameraStatusListener: OnChangedGifCameraStatusListener
    lateinit var onGifCameraTimerCountDownListener: OnGifCameraTimerCountDownListener

    private var gifTimer: GifCameraTimer
    var time: CameraConst.Timer = CameraConst.Timer.ZERO_IN_SECOND
    var imageRotationDegree = 0

    init {
        onChangedGifFrameExtractListener = OnGifFrameExtractListener()
        gifFrameExtractor = GifFrameExtractor(onChangedGifFrameExtractListener, yukiCameraEffectService)
        gifTimer = GifCameraTimer()
    }

    fun registerGifCameraStatusListener(onChangedGifCameraStatusListener: OnChangedGifCameraStatusListener) {
        this.onChangedGifCameraStatusListener = onChangedGifCameraStatusListener
    }

    fun registerGifCameraCountDownListener(onGifCameraTimerCountDownListener: OnGifCameraTimerCountDownListener) {
        this.onGifCameraTimerCountDownListener = onGifCameraTimerCountDownListener
        gifTimer.registerCountDownListener(OnGifTimerCountDownListener())
    }

    fun startRecordGif(isShortShot: Boolean) {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "startRecordGif() isShortShot : " + isShortShot)
        }
        val limitDuration = if (isShortShot) DEFAULT_RECORD_GIF_DURATION_MS else LONG_RECORD_GIF_DURATION_MS

        gifFrameExtractor?.limitDuration = limitDuration

        if (time.seconds > 0) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "startRecordGif() time : " + time.seconds.toLong())
            }
            gifTimer.startTimer(time.seconds.toLong())
            return
        }

        if (gifTimer.isRunning() || isRecordingStarted()) {
            return
        }

        updateGifCameraStatus(if (isShortShot) RecordGifState.START_SHORT_SHOT else RecordGifState.START_LONG_SHOT)
        gifFrameExtractor?.startFrameExtract(imageRotationDegree)
    }

    fun stopRecordGif() {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "stopRecordGif()")
        }
        gifFrameExtractor?.stopFrameExtract()
    }

    fun cancelRecordGif() {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "cancelRecordGif()")
        }
        if (!requestCancelTimer()) {
            gifFrameExtractor?.cancelFrameExtract()
        }
    }

    fun updateGifCameraStatus(gifRecordStatus: RecordGifState) {
        this.gifRecordStatus = gifRecordStatus
        onChangedGifCameraStatusListener.onChangeGifCameraRecordingStatus(gifRecordStatus)
    }

    fun requestCancelTimer(): Boolean {
        if (!gifTimer.isRunning()) {
            return false
        }
        gifTimer.requestCancel()
        return true
    }

    private fun isRecordingStarted(): Boolean {
        return gifRecordStatus == RecordGifState.RECORDING || gifRecordStatus == RecordGifState.START_SHORT_SHOT || gifRecordStatus == RecordGifState.START_LONG_SHOT
    }

    inner class OnGifFrameExtractListener : OnChangedGifFrameExtractListener {
        override fun onCanceledExtractFrame() = updateGifCameraStatus(RecordGifState.READY)

        override fun onFrameExtractStarted() = updateGifCameraStatus(RecordGifState.RECORDING)

        override fun onCompleteSaveFile(dstFileDir: String, delayFrame: Long) {
            updateGifCameraStatus(RecordGifState.STOP)
            onChangedGifCameraStatusListener.onCompleteSaveFiles(dstFileDir, delayFrame)
        }

        override fun onCompleteExtractFrame() {
            updateGifCameraStatus(RecordGifState.STOP)
            onChangedGifCameraStatusListener.onCompleteRecording()
        }

        override fun onError(errorCode: Int) = onChangedGifCameraStatusListener.onError(errorCode)
    }

    inner class OnGifTimerCountDownListener : OnGifCameraTimerCountDownListener {
        override fun onCountDownCanceled() = onGifCameraTimerCountDownListener.onCountDownCanceled()

        override fun onCountDown(seconds: Int) =
                onGifCameraTimerCountDownListener.onCountDown(seconds)

        override fun onCompleteCountDown() {
            onGifCameraTimerCountDownListener.onCompleteCountDown()
            updateGifCameraStatus(RecordGifState.START_SHORT_SHOT)
            gifFrameExtractor?.limitDuration = DEFAULT_RECORD_GIF_DURATION_MS
            gifFrameExtractor?.startFrameExtract(imageRotationDegree)
        }
    }
}
