package com.linecorp.line.camera.util

import android.app.Activity
import android.view.View
import android.view.View.GONE
import android.view.View.VISIBLE
import com.linecorp.line.camera.util.StickerFilterAnimationUtils.Companion.ALPHA_GONE_VALUE
import com.linecorp.line.camera.util.StickerFilterAnimationUtils.Companion.ALPHA_VISIBLE_VALUE
import com.linecorp.line.camera.util.StickerFilterAnimationUtils.Companion.ANIMATION_ALPHA_DURATION_MS
import com.linecorp.line.camera.util.StickerFilterAnimationUtils.Companion.ANIMATION_DURATION_MS
import com.linecorp.line.camera.util.StickerFilterAnimationUtils.Companion.RECORD_BUTTON_DEFAULT_SIZE_PERCENT
import com.linecorp.line.camera.util.StickerFilterAnimationUtils.Companion.RECORD_BUTTON_SMALL_SIZE_PERCENT

class StickerFilterListAnimationUpdater {
    private val TAG = "StickerFilterListAnimationUpdater"

    private val activity: Activity
    private lateinit var bottomContainerView: View
    private lateinit var recordView: View

    private var faceStickerNullCheckView: View? = null // for checking null
    private lateinit var faceStickerContainerView: View
    private lateinit var faceStickerViewList: View
    private lateinit var faceStickerPaddingView: View

    private var filterNullCheckView: View? = null // for checking null
    private lateinit var filterListContainerView: View
    private lateinit var filterListView: View
    private lateinit var filterListPaddingView: View

    private var relatedViews = ArrayList<View>()
    private var stickerFilterAnimationUtils: StickerFilterAnimationUtils

    private val cameraContentPaddingAnimation: CameraContentResizeAnimation = CameraContentResizeAnimation().apply {
        duration = ANIMATION_DURATION_MS
    }

    private var isAnimationRunning: Boolean = false

    constructor(activity: Activity) {
        this.activity = activity
        this.stickerFilterAnimationUtils = StickerFilterAnimationUtils(activity)
    }

    fun setRecordingView(view: View): StickerFilterListAnimationUpdater {
        this.recordView = view
        return this
    }

    fun setBottomContainerView(view: View): StickerFilterListAnimationUpdater {
        this.bottomContainerView = view
        return this
    }

    fun setFilterListView(view: View): StickerFilterListAnimationUpdater {
        this.filterListView = view
        this.filterNullCheckView = view
        return this
    }

    fun setFilterListPaddingView(view: View): StickerFilterListAnimationUpdater {
        this.filterListPaddingView = view
        return this
    }

    fun setFilterListContainerView(view: View): StickerFilterListAnimationUpdater {
        this.filterListContainerView = view
        return this
    }

    fun setFaceStickerContainerView(view: View): StickerFilterListAnimationUpdater {
        this.faceStickerContainerView = view
        return this
    }

    fun setFaceStickerListView(view: View): StickerFilterListAnimationUpdater {
        this.faceStickerNullCheckView = view
        this.faceStickerViewList = view
        return this
    }

    fun setFaceStickerListPaddingView(view: View): StickerFilterListAnimationUpdater {
        this.faceStickerPaddingView = view
        return this
    }

    fun setRelatedViews(view: View): StickerFilterListAnimationUpdater {
        relatedViews.add(view)
        return this
    }

    fun showFilterView() {
        if (isAnimationRunning) {
            return
        }
        if (hideFaceStickerViewIfVisible()) {
            showViewAnimation(filterListView)
        } else {
            showListViewWithRecordingViewAnim(filterListView)
        }
        hideRelatedView()
        showPaddingView(filterListPaddingView, stickerFilterAnimationUtils.defaultPaddingHeight, stickerFilterAnimationUtils.filterPaddingHeight)
        showViewAnimation(filterListContainerView)
    }

    fun hideFilterViewWithRecordingViewAnim() {
        if (isAnimationRunning) {
            return
        }
        isAnimationRunning = true

        showRelatedView()
        hideListViewWithRecordingViewAnim(filterListView)
        hideViewAnimation(filterListContainerView)
        hidePaddingView(filterListPaddingView, stickerFilterAnimationUtils.filterPaddingHeight, stickerFilterAnimationUtils.defaultPaddingHeight)
    }

    fun showStickerView() {
        if (isAnimationRunning) {
            return
        }
        if (hideFilterListViewIfVisible()) {
            showViewAnimation(faceStickerViewList)
        } else {
            showListViewWithRecordingViewAnim(faceStickerViewList)
        }

        hideRelatedView()
        showPaddingView(faceStickerPaddingView, stickerFilterAnimationUtils.defaultPaddingHeight, stickerFilterAnimationUtils.stickerPaddingHeight)
        showViewAnimation(faceStickerContainerView)
    }

    fun hideStickerViewWithRecordingViewAnim() {
        if (isAnimationRunning) {
            return
        }
        isAnimationRunning = true

        showRelatedView()
        hideListViewWithRecordingViewAnim(faceStickerViewList)
        hidePaddingView(faceStickerPaddingView, stickerFilterAnimationUtils.stickerPaddingHeight, stickerFilterAnimationUtils.defaultPaddingHeight)
        hideViewAnimation(faceStickerContainerView)
    }

    fun hideListIfVisible() {
        if (faceStickerNullCheckView != null && faceStickerContainerView.visibility == VISIBLE) {
            hideStickerViewWithRecordingViewAnim()
        } else if (filterNullCheckView != null && filterListContainerView.visibility == VISIBLE) {
            hideFilterViewWithRecordingViewAnim()
        }
        showRelatedView()
    }

    private fun hideFilterListViewIfVisible(): Boolean {
        if (filterNullCheckView != null && filterListContainerView.visibility == VISIBLE) {
            hideViewAnimation(filterListView, false)
            hideViewAnimation(filterListContainerView, false)
            return true
        }
        return false
    }

    private fun hideFaceStickerViewIfVisible(): Boolean {
        if (faceStickerNullCheckView != null && faceStickerContainerView.visibility == VISIBLE) {
            hideViewAnimation(faceStickerViewList, false)
            hideViewAnimation(faceStickerContainerView, false)
            return true
        }
        return false
    }

    private fun showViewAnimation(view: View) {
        view.alpha = 0f
        view.visibility = VISIBLE
        view.animate().cancel()
        view.animate().alpha(ALPHA_VISIBLE_VALUE).duration = ANIMATION_ALPHA_DURATION_MS
    }

    private fun hideViewAnimation(view: View, withAnim: Boolean = true) {
        view.animate().cancel()
        if (withAnim) {
            view.animate().alpha(ALPHA_GONE_VALUE).setDuration(ANIMATION_ALPHA_DURATION_MS).withEndAction(GoneAction(view))
        } else {
            view.visibility = GONE
        }
    }

    private fun showListViewWithRecordingViewAnim(view: View) {
        // make the recording view to be small and move down the bottom view
        makeSmallSizeRecordingView()

        // show the list view
        showViewAnimation(view)
    }

    private fun hideListViewWithRecordingViewAnim(view: View) {
        // make the record to be the default size view and move up the bottom view
        makeDefaultSizeRecordingView()

        // hide the list view
        hideViewAnimation(view)
    }

    private fun makeSmallSizeRecordingView() {
        bottomContainerView.animate().cancel()
        bottomContainerView.animate().translationY(stickerFilterAnimationUtils.recordingContainerDownOffset).duration = ANIMATION_DURATION_MS

        recordView.animate().cancel()
        recordView.animate().scaleX(RECORD_BUTTON_SMALL_SIZE_PERCENT).scaleY(RECORD_BUTTON_SMALL_SIZE_PERCENT).duration = ANIMATION_DURATION_MS
    }

    private fun makeDefaultSizeRecordingView() {
        bottomContainerView.animate().cancel()
        bottomContainerView.animate().translationY(stickerFilterAnimationUtils.recordingContainerUpOffset).duration = ANIMATION_DURATION_MS

        recordView.animate().cancel()
        recordView.animate().scaleX(RECORD_BUTTON_DEFAULT_SIZE_PERCENT).scaleY(RECORD_BUTTON_DEFAULT_SIZE_PERCENT).duration = ANIMATION_DURATION_MS
    }

    private fun showPaddingView(view: View, startOffset: Int, endOffset: Int) {
        showViewAnimation(view)

        cameraContentPaddingAnimation.apply {
            this.view = view
            startHeight = startOffset
            deltaHeight = endOffset - startHeight
        }
        view.startAnimation(cameraContentPaddingAnimation)
    }

    private fun hidePaddingView(view: View, startOffset: Int, endOffset: Int) {
        hideViewAnimation(view)

        cameraContentPaddingAnimation.apply {
            this.view = view
            startHeight = startOffset
            deltaHeight = endOffset - startHeight
        }

        view.startAnimation(cameraContentPaddingAnimation)
    }

    private fun hideRelatedView() {
        relatedViews.map { view -> view.visibility = GONE }
    }

    private fun showRelatedView() {
        relatedViews.map { view -> view.visibility = VISIBLE }
    }

    private inner class GoneAction(internal var container: View) : Runnable {
        override fun run() {
            container.visibility = GONE
            // This gone action have a duration. When the view is not gone yet, the event for the visible animation can occur.
            isAnimationRunning = false
        }
    }
}
