package com.linecorp.line.camera.view.record

import android.graphics.Bitmap
import android.graphics.Canvas
import android.graphics.Color
import android.graphics.Paint
import android.graphics.PorterDuff
import android.graphics.PorterDuffXfermode
import android.graphics.RectF
import android.util.Log
import android.view.View
import android.view.animation.AnimationUtils
import com.linecorp.line.camera.mode.gif.GifCameraManager
import com.linecorp.line.camera.view.EaseCycleInterpolator
import com.linecorp.line.camera.view.record.DrawRecordingViewUtils.Companion.DASH_SHUTTER_ARC_PADDING_RIGHT
import com.linecorp.line.camera.view.record.DrawRecordingViewUtils.Companion.DASH_SHUTTER_ARC_WIDTH
import com.linecorp.line.camera.view.record.DrawRecordingViewUtils.Companion.LINE_DASH_CIRCLE_WIDTH
import com.linecorp.line.media.BuildConfig

class RecordingAnimationViewController {
    companion object {
        private const val TAG = "RecordingAnimationViewController"
    }

    private val view: View
    private var recordingAnimation: RecordingAnimation

    private var drawRecordingViewUtils: DrawRecordingViewUtils
    private var circlePaint: Paint
    private var circleStrokePaint: Paint
    private val circleRectF = RectF()
    private var rectPaint: Paint? = null
    private var rectStrokePaint: Paint? = null
    private var progressCircleBgPaint: Paint? = null
    private var progressPaint: Paint? = null

    private lateinit var interpolator: EaseCycleInterpolator

    private var viewMode: RecordingViewMode.ViewMode = RecordingViewMode.ViewMode.DEFAULT_SHOOT_CIRCLE_NO_ANIM
    private var viewColorMode: RecordingViewMode.ViewColor = RecordingViewMode.ViewColor.WHITE

    private var w: Int = 0
    private var h: Int = 0

    private var widthWithPadding: Int = 0
    private var heightWithPadding: Int = 0

    private var startTime: Long
    private var isAnimating = true

    constructor(view: View) {
        this.view = view
        this.startTime = AnimationUtils.currentAnimationTimeMillis()
        this.recordingAnimation = getAnimationFactory(viewMode)

        drawRecordingViewUtils = DrawRecordingViewUtils()
        circlePaint = drawRecordingViewUtils.createCirclePaint()
        circleStrokePaint = drawRecordingViewUtils.createCircleStrokePaint()
    }

    fun startAnimation(viewMode: RecordingViewMode.ViewMode) {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "startAnimation: " + viewMode)
        }
        isAnimating = true
        if (this.viewMode === viewMode) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "startAnimation: same request" + viewMode + ", " + this.viewMode)
            }
            return
        }

        this.viewMode = viewMode
        recordingAnimation = getAnimationFactory(viewMode)
        recordingAnimation.updateColor(viewColorMode)

        view.invalidate()
    }

    fun onSizeUpdate(w: Int, h: Int) {
        this.widthWithPadding = w
        this.heightWithPadding = h

        this.w = w - DrawRecordingViewUtils.CIRCLE_WIDTH_PADDING.toInt()
        this.h = h - DrawRecordingViewUtils.CIRCLE_WIDTH_PADDING.toInt()
        val padding = DrawRecordingViewUtils.CIRCLE_WIDTH_PADDING / 2 // because of exceed width with stroke width of paint

        circleRectF.set(padding, padding, w.toFloat() - padding, h.toFloat() - padding)
    }

    fun updateColor(viewColor: RecordingViewMode.ViewColor) {
        this.viewColorMode = viewColor
        this.isAnimating = true

        recordingAnimation.updateColor(viewColor)
        view.postInvalidate()
    }

    fun clearRecordingAnimation() {
        isAnimating = false
    }

    fun onDraw(canvas: Canvas?, width: Float, height: Float) {
        if (BuildConfig.DEBUG) {
            Log.d(TAG, "onDraw: " + viewMode)
        }
        if (canvas == null) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onDraw: canvas null")
            }
            return
        }

        recordingAnimation.onDraw(canvas, width, height)

        if (!isAnimating) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onDraw: isAnimating" + isAnimating)
            }
            isAnimating = false
            return
        }

        if (recordingAnimation.isOverLimit()) {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "onDraw: overLimit : " + recordingAnimation.isOverLimit())
            }
            return
        }

        recordingAnimation.updateSweepAngleOffset()
        view.postInvalidateDelayed(recordingAnimation.getValueOnDrawDelay().toLong())
    }

    private fun getAnimationFactory(viewMode: RecordingViewMode.ViewMode): RecordingAnimation {
        return when (viewMode) {
            RecordingViewMode.ViewMode.DEFAULT_SHOOT_CIRCLE_NO_ANIM -> DefaultAnimation()
            RecordingViewMode.ViewMode.CHANGED_GIF_MODE -> ChangedModeGifAnimation()
            RecordingViewMode.ViewMode.CHANGED_PHOTO_VIDEO_MODE -> ChangedModePhotoAnimation()
            RecordingViewMode.ViewMode.RECORDING_SHORT_SHOT_GIF, RecordingViewMode.ViewMode.RECORDING_LONG_SHOT_GIF -> RecordingGifAnimation()
            RecordingViewMode.ViewMode.RECORDING_VIDEO_PRESSED, RecordingViewMode.ViewMode.RECORDING_VIDEO_NON_PRESSED -> RecordingVideoAnimation()
            RecordingViewMode.ViewMode.GIF_MODE_NO_ANIM -> GifModeNoAnim()
            RecordingViewMode.ViewMode.STOP_SHOT_GIF -> StopRecordingGifAnimation()
        }
    }

    inner class GifModeNoAnim : RecordingAnimation() {
        override fun updateColor(viewColor: RecordingViewMode.ViewColor) {
            circleStrokePaint.color = drawRecordingViewUtils.getShootCircleShadowColor(viewColor)
            circlePaint.color = drawRecordingViewUtils.getShootCircleColor(viewColor)
        }

        override fun onDraw(canvas: Canvas?, width: Float, height: Float) {
            super.onDraw(canvas, width, height)

            drawRecordingViewUtils.drawDashArcsNoAnim(canvas, circleRectF, circleStrokePaint)
            drawRecordingViewUtils.drawDashArcsNoAnim(canvas, circleRectF, circlePaint)
        }
    }

    inner class StopRecordingGifAnimation : RecordingAnimation {
        constructor() {
            if (rectPaint == null) {
                rectPaint = drawRecordingViewUtils.createRectPaint()
            }
            if (rectStrokePaint == null) {
                rectStrokePaint = drawRecordingViewUtils.createRectStrokePaint()
            }
        }

        override fun updateColor(viewColor: RecordingViewMode.ViewColor) {
            circleStrokePaint.color = drawRecordingViewUtils.getShootCircleShadowColor(viewColor)
            circlePaint.color = drawRecordingViewUtils.getShootCircleColor(viewColor)
            rectPaint?.color = drawRecordingViewUtils.getShootRectInsideCircle(viewMode, viewColor)
            rectStrokePaint?.color = drawRecordingViewUtils.getShootRectInsideCircleShadow(viewMode, viewColor)
        }

        override fun onDraw(canvas: Canvas?, width: Float, height: Float) {
            super.onDraw(canvas, width, height)

            drawRecordingViewUtils.drawDashArcsNoAnim(canvas, circleRectF, circleStrokePaint)
            drawRecordingViewUtils.drawDashArcsNoAnim(canvas, circleRectF, circlePaint)

            drawRecordingViewUtils.drawRect(canvas, w, h, rectPaint, rectStrokePaint)
        }
    }

    inner class RecordingGifAnimation : RecordingAnimation {
        var isShort = false
        var limitDuration = 0f

        constructor() {
            if (progressCircleBgPaint == null) {
                progressCircleBgPaint = drawRecordingViewUtils.createProgressCircleBgPaint()
            }

            if (progressPaint == null) {
                progressPaint = drawRecordingViewUtils.createProgressPaint()
            }

            isShort = (viewMode === RecordingViewMode.ViewMode.RECORDING_SHORT_SHOT_GIF)
            limitDuration = if (isShort) GifCameraManager.DEFAULT_RECORD_GIF_DURATION_MS.toFloat() else GifCameraManager.LONG_RECORD_GIF_DURATION_MS.toFloat()
            limitDuration = limitDuration
            startTime = AnimationUtils.currentAnimationTimeMillis()
        }

        override fun onDraw(canvas: Canvas?, width: Float, height: Float) {
            canvas?.drawArc(circleRectF, 0f, 360f, false, progressCircleBgPaint)
            canvas?.drawArc(circleRectF, 270f, sweepAngleOffset, false, progressPaint)
        }

        override fun isOverLimit(): Boolean {
            return sweepAngleOffset > 360
        }

        override fun updateSweepAngleOffset() {
            sweepAngleOffset = 360 * ((AnimationUtils.currentAnimationTimeMillis() - startTime) / limitDuration)
        }

        override fun getValueOnDrawDelay(): Int {
            return 30
        }

        override fun updateColor(viewColor: RecordingViewMode.ViewColor) {
            progressCircleBgPaint?.color = drawRecordingViewUtils.getProgressCircleBgColor(viewColor)
            progressPaint?.color = drawRecordingViewUtils.getProgressCircleColor(viewColor)
        }
    }

    private inner class RecordingVideoAnimation : RecordingAnimation {
        private var bitmapCanvas: Canvas
        private var bitmap: Bitmap? = null
        private var drawPaint: Paint
        private lateinit var clearPaint: Paint

        constructor() {
            interpolator = EaseCycleInterpolator(0.5f)
            if (rectPaint == null) {
                rectPaint = drawRecordingViewUtils.createRectPaint()
            }
            if (rectStrokePaint == null) {
                rectStrokePaint = drawRecordingViewUtils.createRectStrokePaint()
            }
            bitmapCanvas = Canvas()
            drawPaint = Paint()
            startTime = AnimationUtils.currentAnimationTimeMillis()
        }

        override fun onDraw(canvas: Canvas?, width: Float, height: Float) {
            super.onDraw(canvas, width, height)

            var t = ((AnimationUtils.currentAnimationTimeMillis() - startTime) % DrawRecordingViewUtils.ANIMATION_DURATION).toFloat()
            t /= DrawRecordingViewUtils.ANIMATION_DURATION.toFloat()
            t = interpolator.getInterpolation(t)

            canvas?.drawColor(Color.TRANSPARENT, PorterDuff.Mode.OVERLAY)

            if (viewMode === RecordingViewMode.ViewMode.RECORDING_VIDEO_PRESSED) {
                drawRecordingOuterCircle(canvas, t)
            }

            canvas?.drawArc(circleRectF, 0f, 360f, false, circleStrokePaint)
            canvas?.drawArc(circleRectF, 0f, 360f, false, circlePaint)

            drawRecordingViewUtils.drawRect(canvas, w, h, rectPaint, rectStrokePaint)
        }

        override fun updateColor(viewColor: RecordingViewMode.ViewColor) {
            circlePaint.color = drawRecordingViewUtils.getShootCircleWithRectColor(viewMode, viewColor)
            circleStrokePaint.color = drawRecordingViewUtils.getShootCircleShadowWithRectColor(viewMode, viewColor)
            rectPaint?.color = drawRecordingViewUtils.getShootRectInsideCircle(viewMode, viewColor)
            rectStrokePaint?.color = drawRecordingViewUtils.getShootRectInsideCircleShadow(viewMode, viewColor)
        }

        private fun drawRecordingOuterCircle(canvas: Canvas?, t: Float) {
            if (widthWithPadding <= 0 || heightWithPadding <= 0) {
                return
            }

            val dstW = (widthWithPadding * DrawRecordingViewUtils.OUTER_CIRCLE_RATIO + 0.5f).toInt()
            val dstH = (heightWithPadding * DrawRecordingViewUtils.OUTER_CIRCLE_RATIO + 0.5f).toInt()
            val offsetX = -(Math.abs(dstW - widthWithPadding) * 0.5f + 0.5f)
            val offsetY = -(Math.abs(dstH - heightWithPadding) * 0.5f + 0.5f)

            checkAndInit(dstW, dstH)

            val cX = dstW / 2f
            val cY = dstH / 2f
            val inRadius = widthWithPadding / 2.1f
            val outRadius = dstW / 2 * t

            bitmapCanvas.drawColor(Color.TRANSPARENT, PorterDuff.Mode.CLEAR)
            bitmapCanvas.drawCircle(cX, cY, outRadius, drawPaint)
            bitmapCanvas.drawCircle(cX, cY, inRadius, clearPaint)

            canvas?.drawBitmap(bitmap, offsetX, offsetY, null)
        }

        private fun checkAndInit(dstW: Int, dstH: Int) {
            if (bitmap == null || bitmap?.width != dstW || bitmap?.height != dstH) {
                bitmap = Bitmap.createBitmap(dstW, dstH, Bitmap.Config.ARGB_8888)
                bitmapCanvas.setBitmap(bitmap)
            }

            drawPaint.color = drawRecordingViewUtils.getCircleInteractColor(viewMode, viewColorMode)
            drawPaint.isAntiAlias = true
            clearPaint = Paint()
            clearPaint.color = Color.TRANSPARENT
            clearPaint.xfermode = PorterDuffXfermode(PorterDuff.Mode.CLEAR)
            clearPaint.isAntiAlias = true
        }

        override fun getValueOnDrawDelay(): Int {
            return 33
        }

        override fun isOverLimit(): Boolean {
            // only draw animation on RECORDING_VIDEO_PRESSED
            return viewMode !== RecordingViewMode.ViewMode.RECORDING_VIDEO_PRESSED
        }
    }

    private inner class DefaultAnimation : RecordingAnimation() {
        override fun onDraw(canvas: Canvas?, width: Float, height: Float) {
            super.onDraw(canvas, width, height)

            val radius = (width - DrawRecordingViewUtils.CIRCLE_WIDTH_PADDING) / 2
            canvas?.drawCircle(width / 2, width / 2, radius, circleStrokePaint)
            canvas?.drawCircle(width / 2, width / 2, radius, circlePaint)
        }

        override fun updateColor(viewColor: RecordingViewMode.ViewColor) {
            circleStrokePaint.color = drawRecordingViewUtils.getShootCircleShadowColor(viewColor)
            circlePaint.color = drawRecordingViewUtils.getShootCircleColor(viewColor)
        }
    }

    private inner class ChangedModePhotoAnimation : RecordingAnimation {
        override var valueOnDrawLimit: Float = (DASH_SHUTTER_ARC_WIDTH + DASH_SHUTTER_ARC_PADDING_RIGHT)

        constructor() {
            sweepAngleOffset = DASH_SHUTTER_ARC_WIDTH
        }

        override fun onDraw(canvas: Canvas?, width: Float, height: Float) {
            super.onDraw(canvas, width, height)

            drawRecordingViewUtils.drawArcs(canvas, circleRectF, LINE_DASH_CIRCLE_WIDTH, DASH_SHUTTER_ARC_WIDTH, DASH_SHUTTER_ARC_WIDTH + DASH_SHUTTER_ARC_PADDING_RIGHT, sweepAngleOffset, circleStrokePaint)
            drawRecordingViewUtils.drawArcs(canvas, circleRectF, LINE_DASH_CIRCLE_WIDTH, DASH_SHUTTER_ARC_WIDTH, DASH_SHUTTER_ARC_WIDTH + DASH_SHUTTER_ARC_PADDING_RIGHT, sweepAngleOffset, circlePaint)
        }

        override fun updateColor(viewColor: RecordingViewMode.ViewColor) {
            circleStrokePaint.color = drawRecordingViewUtils.getShootCircleShadowColor(viewColor)
            circlePaint.color = drawRecordingViewUtils.getShootCircleColor(viewColor)
        }

        override fun isOverLimit(): Boolean {
            return sweepAngleOffset >= valueOnDrawLimit
        }

        override fun updateSweepAngleOffset() {
            sweepAngleOffset += 1f
        }
    }

    private inner class ChangedModeGifAnimation : RecordingAnimation {
        override var valueOnDrawLimit: Float = DASH_SHUTTER_ARC_WIDTH
            get() = DASH_SHUTTER_ARC_WIDTH

        constructor() {
            sweepAngleOffset = (DASH_SHUTTER_ARC_WIDTH + DASH_SHUTTER_ARC_PADDING_RIGHT)
        }

        override fun updateColor(viewColor: RecordingViewMode.ViewColor) {
            circleStrokePaint.color = drawRecordingViewUtils.getShootCircleShadowColor(viewColor)
            circlePaint.color = drawRecordingViewUtils.getShootCircleColor(viewColor)
        }

        override fun onDraw(canvas: Canvas?, width: Float, height: Float) {
            super.onDraw(canvas, width, height)

            drawRecordingViewUtils.drawDashArcs(canvas, circleRectF, LINE_DASH_CIRCLE_WIDTH, DASH_SHUTTER_ARC_WIDTH + DASH_SHUTTER_ARC_PADDING_RIGHT, sweepAngleOffset, circleStrokePaint)
            drawRecordingViewUtils.drawDashArcs(canvas, circleRectF, LINE_DASH_CIRCLE_WIDTH, DASH_SHUTTER_ARC_WIDTH + DASH_SHUTTER_ARC_PADDING_RIGHT, sweepAngleOffset, circlePaint)
        }

        override fun isOverLimit(): Boolean {
            return sweepAngleOffset < valueOnDrawLimit
        }

        override fun updateSweepAngleOffset() {
            if (BuildConfig.DEBUG) {
                Log.d(TAG, "updateSweepAngleOffset: " + sweepAngleOffset)
            }
            sweepAngleOffset -= 1f
        }
    }
}