package jp.naver.line.androie.common.lib.util

import java.util.regex.Pattern

/**
 * Contains Line app link patterns.
 *
 * Private fields are copy/paste from AppCompat because we can't access them
 * and sadly Android is not allowing us to use the same smoothly custom scheme links with public API.
 * TODO: we should stop ASAP to copy things from Android or at least be sure to update them when original fields are changed.
 **/
object LinePatterns {

    @JvmField
    val LINE_PATTERN = uriPatternWithCustomSchemes(listOf("line", "lineb"))

    @JvmField
    val LINE_FAMILY_APPS_PATTERN = uriPatternWithCustomSchemes(
            AppSchemes.LineFamilyApps.values().map { it.scheme })

    @Suppress("DEPRECATION")
    @JvmField
    val NHNCORP_APPS_PATTERN = uriPatternWithCustomSchemes(
            // Filter :// suffix and keep only custom schemes
            AppSchemes.NHNCORP_APP_SCHEMES.map { it.substring(0, it.length - 3) })

    /**
     * Valid UCS characters defined in RFC 3987. Excludes space characters.
     * @see android.support.v4.util.PatternsCompat.UCS_CHAR
     */
    private const val UCS_CHAR = "[" +
            "\u00A0-\uD7FF" +
            "\uF900-\uFDCF" +
            "\uFDF0-\uFFEF" +
            "\uD800\uDC00-\uD83F\uDFFD" +
            "\uD840\uDC00-\uD87F\uDFFD" +
            "\uD880\uDC00-\uD8BF\uDFFD" +
            "\uD8C0\uDC00-\uD8FF\uDFFD" +
            "\uD900\uDC00-\uD93F\uDFFD" +
            "\uD940\uDC00-\uD97F\uDFFD" +
            "\uD980\uDC00-\uD9BF\uDFFD" +
            "\uD9C0\uDC00-\uD9FF\uDFFD" +
            "\uDA00\uDC00-\uDA3F\uDFFD" +
            "\uDA40\uDC00-\uDA7F\uDFFD" +
            "\uDA80\uDC00-\uDABF\uDFFD" +
            "\uDAC0\uDC00-\uDAFF\uDFFD" +
            "\uDB00\uDC00-\uDB3F\uDFFD" +
            "\uDB44\uDC00-\uDB7F\uDFFD" +
            "&&[^\u00A0[\u2000-\u200A]\u2028\u2029\u202F\u3000]]"

    /**
     * Valid characters for IRI label defined in RFC 3987.
     * @see android.support.v4.util.PatternsCompat.LABEL_CHAR
     */
    private const val LABEL_CHAR = "a-zA-Z0-9" + UCS_CHAR

    /**
     * Match scheme intent separator (://) followed by path and query.
     * @see android.support.v4.util.PatternsCompat.PATH_AND_QUERY
     */
    private const val INTENT_PATH_AND_QUERY = ("://(?:(?:[" + LABEL_CHAR
            + ";/\\?:@&=#~" // plus optional query params
            + "\\-\\.\\+!\\*'\\(\\),_\\$])|(?:%[a-fA-F0-9]{2}))*")

    @JvmStatic
    private fun uriPatternWithCustomSchemes(schemes: List<String>) =
            Pattern.compile(
                    "(" + schemes.joinToString(separator = "|") + ")" + INTENT_PATH_AND_QUERY,
                    Pattern.CASE_INSENSITIVE)
}
