package jp.naver.line.androie.common.theme

import android.content.res.ColorStateList
import android.graphics.Color
import android.graphics.PorterDuff
import android.graphics.drawable.ColorDrawable
import android.graphics.drawable.StateListDrawable
import android.support.annotation.ColorInt
import android.support.annotation.VisibleForTesting
import android.support.v4.graphics.drawable.DrawableCompat
import android.support.v4.view.TintableBackgroundView
import android.support.v4.view.ViewCompat
import android.support.v4.widget.ImageViewCompat
import android.support.v7.widget.AppCompatImageView
import android.view.View
import android.widget.ImageView
import android.widget.TextView
import jp.naver.android.common.R
import jp.naver.line.androie.common.view.ImageTintable

/**
 * Model to represent color which is used by the LINE theme. This has at least one of
 * [ColorStateList] or [ColorInt].
 */
// TODO: Make this sealed class.
// TODO: Use strategy pattern
class ThemeElementColorData(
        /**
         * Returns the raw `ColorStateList` value specified in the theme JSON.
         * If the value is specified as a single color value, returns `null`.
         */
        @VisibleForTesting
        val rawColorStateList: ColorStateList?,
        private val color: Int?
) {
    /**
     * Returns the `ColorStateList` for the this Theme element.
     * If the value is specified as a single color value in the theme JSON,
     * the returned `ColorStateList` only contains the color.
     */
    val colorStateList: ColorStateList
        get() = rawColorStateList ?: ColorStateList.valueOf(colorInt)

    val colorInt: Int
        @ColorInt get() = color ?: rawColorStateList?.defaultColor ?: Color.BLACK

    fun applyToImageTint(view: View): Boolean {
        when (view) {
            is ImageTintable -> view.tintImage(colorStateList)
            is AppCompatImageView -> ImageViewCompat.setImageTintList(view, colorStateList)
            is ImageView -> view.setColorFilter(colorInt, PorterDuff.Mode.SRC_IN)
        }
        // TODO: We may need to return false when view is not ImageView or ImageTintable.
        return true
    }

    fun applyToBackground(view: View): Boolean {
        if (rawColorStateList != null) {
            applyBackgroundColorState(view, rawColorStateList)
        } else {
            view.setBackgroundColor(colorInt)
        }
        return true
    }

    fun applyToBackgroundTint(view: View): Boolean {
        val background = view.background ?: return false
        when {
            view is TintableBackgroundView -> ViewCompat.setBackgroundTintList(view, colorStateList)
            background is ColorDrawable -> background.color = colorInt
            else -> view.background = DrawableCompat.wrap(background).mutate()
                    .also { DrawableCompat.setTintList(it, colorStateList) }
        }
        return true
    }

    fun applyToText(view: TextView): Boolean {
        if (rawColorStateList != null) {
            view.setTextColor(rawColorStateList)
        } else {
            view.setTextColor(colorInt)
        }
        return true
    }

    fun applyToTextShadow(view: TextView): Boolean {
        val shadowThicknessDp = view.resources
                .getDimensionPixelSize(R.dimen.theme_text_shadow_thickness)
                .toFloat()
        view.setShadowLayer(shadowThicknessDp, 0f, shadowThicknessDp, colorInt)
        return true
    }

    fun applyToHintText(view: TextView): Boolean {
        view.setHintTextColor(colorInt)
        return true
    }

    override fun toString(): String = "color{color=$color, colors=$rawColorStateList}"

    companion object {
        private val DEFAULT_COLOR_STATE = intArrayOf()
        private val BACKGROUND_COLOR_STATES = arrayOf(
                intArrayOf(android.R.attr.state_pressed),
                intArrayOf(android.R.attr.state_selected),
                intArrayOf(-android.R.attr.state_enabled))

        private fun applyBackgroundColorState(view: View, colorStateList: ColorStateList) {
            val drawable = StateListDrawable()
            val defaultColorInt = colorStateList.defaultColor
            BACKGROUND_COLOR_STATES.map { state ->
                state to colorStateList.getColorForState(state, defaultColorInt)
            }.filter { (_, colorInt) ->
                colorInt != defaultColorInt
            }.forEach { (state, colorInt) ->
                drawable.addState(state, ColorDrawable(colorInt))
            }
            drawable.addState(DEFAULT_COLOR_STATE, ColorDrawable(defaultColorInt))
            view.background = drawable
        }
    }
}

fun ThemeElementColorData?.applyNullableToImageTint(view: View): Boolean =
        this?.let { applyToImageTint(view) } ?: false

fun ThemeElementColorData?.applyNullableToBackground(view: View): Boolean =
        this?.let { applyToBackground(view) } ?: false

fun ThemeElementColorData?.applyNullableToBackgroundTint(view: View): Boolean =
        this?.let { applyToBackgroundTint(view) } ?: false

fun ThemeElementColorData?.applyNullableToText(view: TextView): Boolean =
        this?.let { applyToText(view) } ?: false

fun ThemeElementColorData?.applyNullableToTextShadow(view: TextView): Boolean =
        this?.let { applyToTextShadow(view) } ?: false

fun ThemeElementColorData?.applyNullableToHintText(view: TextView): Boolean =
        this?.let { applyToHintText(view) } ?: false
