package jp.naver.line.androie.common.theme

import android.content.res.Resources
import android.graphics.Rect
import android.support.annotation.ColorRes
import android.support.annotation.DrawableRes
import android.support.v4.content.ContextCompat
import android.view.View
import android.widget.ImageView
import android.widget.TextView
import jp.naver.line.androie.common.theme.ViewPaddingExtensions.padding

/**
 * Data class to represent the default value of a theme element.
 */
data class ThemeElementDefaultValue(
        val type: ValueType,
        val defaultValueResourceId: Int
) {

    fun applyDefaultTheme(view: View) = when (type) {
        ValueType.BACKGROUND_DRAWABLE ->
            applyDefaultThemeAsBackgroundDrawable(view, defaultValueResourceId)
        ValueType.BACKGROUND_COLOR ->
            applyDefaultThemeAsBackgroundColor(view, defaultValueResourceId)
        ValueType.IMAGE_DRAWABLE ->
            applyDefaultThemeAsImageDrawable(view, defaultValueResourceId)
        ValueType.TEXT_COLOR ->
            applyDefaultThemeAsTextColor(view, defaultValueResourceId)
    }

    private fun applyDefaultThemeAsBackgroundDrawable(view: View, @DrawableRes resourceId: Int) {
        val drawable = ContextCompat.getDrawable(view.context, resourceId) ?: return
        val padding = view.padding
        view.background = drawable
        if (padding.hasValue) {
            view.padding = padding
        }
    }

    private fun applyDefaultThemeAsBackgroundColor(view: View, @ColorRes resourceId: Int) =
            try {
                view.setBackgroundColor(ContextCompat.getColor(view.context, resourceId))
            } catch (e: Resources.NotFoundException) {
                ThemeErrorReportUtil.report(
                        "Error while getting color $resourceId",
                        "ThemeManager.applyDefault(View, int, int)",
                        e)
            }

    private fun applyDefaultThemeAsImageDrawable(view: View, @DrawableRes resourceId: Int) =
            ContextCompat.getDrawable(view.context, resourceId)?.let { drawable ->
                (view as? ImageView)?.setImageDrawable(drawable)
            }

    private fun applyDefaultThemeAsTextColor(view: View, @ColorRes resourceId: Int) =
            try {
                (view as? TextView)
                        ?.setTextColor(ContextCompat.getColorStateList(view.context, resourceId))
            } catch (e: Resources.NotFoundException) {
                ThemeErrorReportUtil.report(
                        "Error while getting text color $resourceId",
                        "ThemeManager.applyDefault(View, int, int)",
                        e)
            }

    private val Rect.hasValue get() = left != 0 || top != 0 || right != 0 || bottom != 0

    /**
     * Enum to represent a type of theme element default value.
     */
    enum class ValueType {
        BACKGROUND_DRAWABLE,
        BACKGROUND_COLOR,
        IMAGE_DRAWABLE,
        TEXT_COLOR,
    }
}
