package jp.naver.line.androie.common.theme

import android.graphics.Rect
import android.graphics.drawable.Drawable
import android.view.View
import android.widget.CompoundButton
import android.widget.ImageView
import jp.naver.line.androie.common.theme.ViewPaddingExtensions.padding
import java.util.Arrays

/**
 * Model to represent image which is used by the LINE theme. The main data is image file names which
 * are obtained by [.getBaseImage] and [.getOverlayImages]. This also keeps
 * additional information such as the tile mode or image padding.
 */
// TODO: Use strategy pattern
class ThemeElementImageData(
        val baseImage: ThemeElementBaseImageData,
        val overlayImages: Array<ThemeElementOverlayImageData>?,
        val isRepeatImage: Boolean,
        val padding: Rect?) {

    fun applyToImage(
            view: View,
            themeResources: ThemeResources
    ): Boolean {
        val drawable = getDrawable(themeResources) ?: return false
        if (view is ImageView) {
            padding?.let { view.padding = it }
            view.setImageDrawable(drawable)
        }
        return true // TODO: We may need to return false when view is not ImageView.
    }

    fun applyToBackground(
            view: View,
            themeResources: ThemeResources
    ): Boolean {
        val drawable = getDrawable(themeResources) ?: return false

        // TODO: Consider to remove CompoundButton case.
        when (view) {
            is CompoundButton -> view.buttonDrawable = drawable
            else -> view.background = drawable
        }

        // Overrides the padding of the given view according to the following rule.
        // 1. If ImageInfo requires no padding explicitly, we set the view padding with 0px.
        // 2. If ImageInfo has other padding setting, we use it.
        // 3. If the given drawable has a padding, we use it. (See, [Drawable.getPadding].)
        // 4. Otherwise, do nothing.
        val newPadding = padding ?: getDrawablePadding(drawable)
        newPadding?.let { view.padding = newPadding }

        return true
    }

    fun getDrawable(themeResources: ThemeResources): Drawable? {
        var retryCount = 0
        while (retryCount <= 1) {
            try {
                return themeResources.getDrawable(this)
            } catch (e: Exception) {
                // When OutOfMemory or something happens, try to reload after clearing the cache.
                // TODO: Use more specific Error.
                themeResources.clearCache()
                ThemeErrorReportUtil.report(
                        "Error while getting drawable $this",
                        "ThemeManager.getDrawable(ImageInfo)",
                        e)
                retryCount++
            }
        }
        return null
    }

    override fun toString(): String {
        return "image{baseImage=$baseImage, " +
                "isRepeatImage=$isRepeatImage, " +
                "overlayImages=${Arrays.toString(overlayImages)}, " +
                "padding=${padding?.toShortString()}}"
    }

    class Builder(imageData: ThemeElementImageData? = null) {
        private var baseImage: ThemeElementBaseImageData? = imageData?.baseImage
        private var overlayImages: Array<ThemeElementOverlayImageData>? = imageData?.overlayImages
        private var isRepeatImage: Boolean = imageData != null && imageData.isRepeatImage
        private var padding: Rect? = imageData?.padding

        fun setBaseImage(baseImage: ThemeElementBaseImageData?): Builder =
                apply { this.baseImage = baseImage }

        fun setOverlayImages(overlayImages: Array<ThemeElementOverlayImageData>?): Builder =
                apply { this.overlayImages = overlayImages }

        fun setRepeatImage(isRepeatImage: Boolean): Builder =
                apply { this.isRepeatImage = isRepeatImage }

        fun setPadding(padding: Rect?): Builder =
                apply { this.padding = padding }

        /**
         * Returns [ThemeElementImageData].
         * If there is not enough data to build it, returns null.
         */
        fun build(): ThemeElementImageData? = baseImage?.let { nonNullBaseImage ->
            ThemeElementImageData(nonNullBaseImage, overlayImages, isRepeatImage, padding)
        }
    }

    companion object {
        private fun getDrawablePadding(drawable: Drawable): Rect? {
            val padding = Rect()
            val isValid = drawable.getPadding(padding)
            return if (isValid) padding else null
        }
    }
}

fun ThemeElementImageData?.applyNullableToBackground(
        view: View,
        themeResources: ThemeResources
): Boolean = this?.let { applyToBackground(view, themeResources) } ?: false

fun ThemeElementImageData?.applyNullableToImage(
        view: View,
        themeResources: ThemeResources
): Boolean = this?.let { applyToImage(view, themeResources) } ?: false
