package jp.naver.line.androie.common.theme

import android.graphics.Rect
import android.view.View
import android.widget.TextView

/**
 * Set of theme colors and images for a corresponding ThemeElementKey (ThemeElementKey is currently
 * a view id). Internally, this holds [ThemeElementImageData] and [ThemeElementColorData] to
 * represent the colors and images.
 *
 *
 * Theme items are as follows.
 *
 *  *  background color
 *  *  background image
 *  *  tint color of background image
 *  *  image for ImageView
 *  *  tint color for ImageView
 *  *  text color for TextView
 *  *  hint text color for TextView
 *
 */
class ThemeElementValue(
        val image: ThemeElementImageData?,
        val imageTintColor: ThemeElementColorData?,
        val backgroundColor: ThemeElementColorData?,
        val backgroundImage: ThemeElementImageData?,
        val backgroundImageTintColor: ThemeElementColorData?,
        val textColor: ThemeElementColorData?,
        val textShadowColor: ThemeElementColorData?,
        val hintTextColor: ThemeElementColorData?) {

    val isEmpty: Boolean = image == null
            && imageTintColor == null
            && backgroundColor == null
            && backgroundImage == null
            && backgroundImageTintColor == null
            && textColor == null
            && textShadowColor == null
            && hintTextColor == null

    fun emptyToNull(): ThemeElementValue? = if (!isEmpty) this else null

    /**
     * Apply this theme element to the given `view`. Then, returns true if some theme value is
     * applied.
     */
    fun applyTheme(
            view: View,
            themeResources: ThemeResources,
            applyingTargets: Collection<ThemeElementValueType>
    ): Boolean = applyingTargets.fold(false) { applied, type ->
        applyThemeForType(view, themeResources, type) || applied
    }

    override fun toString(): String {
        val fieldStrings = mutableListOf<String>()

        image?.let { fieldStrings.add("image=" + it) }
        imageTintColor?.let { fieldStrings.add("imageTintColor=" + it) }
        backgroundColor?.let { fieldStrings.add("backgroundColor=" + it) }
        backgroundImage?.let { fieldStrings.add("backgroundImage=" + it) }
        backgroundImageTintColor?.let { fieldStrings.add("backgroundImageTintColor=" + it) }
        textColor?.let { fieldStrings.add("textColor=" + it) }
        hintTextColor?.let { fieldStrings.add("hintTextColor=" + it) }
        textShadowColor?.let { fieldStrings.add("textShadowColor=" + it) }
        fieldStrings.add("isEmpty=" + isEmpty)

        return fieldStrings.joinToString(prefix = "ThemeItemInfo{", postfix = "}")
    }

    private fun applyThemeForType(
            view: View,
            themeResources: ThemeResources,
            type: ThemeElementValueType
    ): Boolean = when (type) {
        ThemeElementValueType.BACKGROUND -> applyToBackground(view, themeResources)
        ThemeElementValueType.IMAGE -> applyToImage(view, themeResources)
        ThemeElementValueType.TEXT -> applyToText(view)
    }

    private fun applyToBackground(
            view: View,
            themeResources: ThemeResources
    ): Boolean = backgroundImage.applyNullableToBackground(view, themeResources) or
            backgroundColor.applyNullableToBackground(view) or
            backgroundImageTintColor.applyNullableToBackgroundTint(view)

    private fun applyToImage(
            view: View,
            themeResources: ThemeResources
    ): Boolean = image.applyNullableToImage(view, themeResources) or
            imageTintColor.applyNullableToImageTint(view)

    private fun applyToText(view: View): Boolean = if (view is TextView) {
        textColor.applyNullableToText(view) or
                textShadowColor.applyNullableToTextShadow(view) or
                hintTextColor.applyNullableToHintText(view)
    } else {
        false
    }

    // TODO: Remove this after applying to all the code.
    enum class ResourceType {
        BG_COLOR,
        TEXT_COLOR,
    }

    // TODO: Remove this
    class Builder {
        private val imageBuilder: ThemeElementImageData.Builder = ThemeElementImageData.Builder()
        private val backgroundImageBuilder: ThemeElementImageData.Builder =
                ThemeElementImageData.Builder()

        private var imageTintColor: ThemeElementColorData? = null
        private var backgroundColor: ThemeElementColorData? = null
        private var backgroundImageTintColor: ThemeElementColorData? = null
        private var textColor: ThemeElementColorData? = null
        private var textShadowColor: ThemeElementColorData? = null
        private var hintTextColor: ThemeElementColorData? = null

        fun setImage(image: ThemeElementImageData?): Builder = also {
            imageBuilder
                    .setBaseImage(image?.baseImage)
                    .setOverlayImages(image?.overlayImages)
                    .setRepeatImage(image != null && image.isRepeatImage)
                    .setPadding(image?.padding)
        }

        fun setBaseImage(baseImage: ThemeElementBaseImageData?): Builder =
                apply { imageBuilder.setBaseImage(baseImage) }

        fun setOverlayImages(overlayImages: Array<ThemeElementOverlayImageData>?): Builder =
                apply { imageBuilder.setOverlayImages(overlayImages) }

        fun setImageRepeat(isRepeatImage: Boolean): Builder =
                apply { imageBuilder.setRepeatImage(isRepeatImage) }

        fun setImagePadding(padding: Rect?): Builder =
                apply { imageBuilder.setPadding(padding) }

        fun setImageTintColor(imageTintColor: ThemeElementColorData?): Builder =
                apply { this.imageTintColor = imageTintColor }

        fun setBackgroundColor(backgroundColor: ThemeElementColorData?): Builder =
                apply { this.backgroundColor = backgroundColor }

        fun setBackgroundImage(backgroundImage: ThemeElementImageData?): Builder = apply {
            backgroundImageBuilder
                    .setBaseImage(backgroundImage?.baseImage)
                    .setOverlayImages(backgroundImage?.overlayImages)
                    .setRepeatImage(backgroundImage != null && backgroundImage.isRepeatImage)
                    .setPadding(backgroundImage?.padding)
        }

        fun setBackgroundBaseImage(baseImage: ThemeElementBaseImageData?): Builder =
                apply { backgroundImageBuilder.setBaseImage(baseImage) }

        fun setBackgroundOverlayImages(
                overlayImages: Array<ThemeElementOverlayImageData>?
        ): Builder = apply { backgroundImageBuilder.setOverlayImages(overlayImages) }

        fun setBackgroundImageRepeat(isRepeatImage: Boolean): Builder =
                apply { backgroundImageBuilder.setRepeatImage(isRepeatImage) }

        fun setBackgroundImagePadding(padding: Rect?): Builder =
                apply { backgroundImageBuilder.setPadding(padding) }

        fun setBackgroundImageTintColor(
                backgroundImageTintColor: ThemeElementColorData?
        ): Builder = apply { this.backgroundImageTintColor = backgroundImageTintColor }

        fun setTextColor(textColor: ThemeElementColorData?): Builder =
                apply { this.textColor = textColor }

        fun setTextShadowColor(textShadowColor: ThemeElementColorData?): Builder =
                apply { this.textShadowColor = textShadowColor }

        fun setHintTextColor(hintTextColor: ThemeElementColorData?): Builder =
                apply { this.hintTextColor = hintTextColor }

        fun build(): ThemeElementValue {
            return ThemeElementValue(
                    imageBuilder.build(),
                    imageTintColor,
                    backgroundColor,
                    backgroundImageBuilder.build(),
                    backgroundImageTintColor,
                    textColor,
                    textShadowColor,
                    hintTextColor)
        }
    }

    companion object {
        val INVALID = Builder().build()
    }
}