package jp.naver.line.androie.common.theme

import com.linecorp.android.security.io.SafeZipInputStream
import java.io.BufferedInputStream
import java.io.File
import java.io.FileInputStream
import java.io.FileOutputStream
import java.io.IOException
import java.util.zip.ZipEntry
import java.util.zip.ZipInputStream

/**
 * A manager of a theme file. This extracts a zipped theme file and copies it to the internal
 * storage. This also deletes unused theme file from internal storage.
 */
abstract class ThemeFileManager {

    abstract val jsonFile: File

    abstract val loadedThemeDirectory: File

    protected abstract val loadedImageDirectory: File

    protected abstract val loadedTimelineImageDirectory: File

    protected abstract val loadedGroupHomeImageDirectory: File

    fun flushThemeLoadFolder() = deleteFilesRecursively(loadedThemeDirectory)

    protected fun maybeCreateDirectory(parent: File, directoryName: String): File =
            File(parent, directoryName).also { directory ->
                if (!directory.exists()) {
                    directory.mkdirs()
                }
            }

    @Throws(IOException::class)
    fun extract(zipFile: File) = SafeZipInputStream(BufferedInputStream(FileInputStream(zipFile)))
            .use { zipInputStream ->
                zipInputStream.forEachNonDirectoryFile { entry ->
                    extract(getDestinationFile(entry.name), zipInputStream)
                }
            }

    private fun getDestinationFile(sourceFilePath: String): File? {
        val sourceFile = File(sourceFilePath)
        val sourcePath = sourceFile.absolutePath

        // Case json file.
        if (sourcePath.endsWith(JSON_FILE_SUFFIX)) {
            return jsonFile
        }

        // Case for other files.
        val lastSeparatorIndex = Math.max(sourcePath.lastIndexOf(PATH_SEPARATOR), 0)
        val parentPath = sourcePath.substring(0, lastSeparatorIndex)
        val destinationParent: File? = when {
            parentPath.endsWith(THEME_IMAGE_DIRECTORY) -> loadedImageDirectory
            parentPath.endsWith(THEME_TIMELINE_DIRECTORY) -> loadedTimelineImageDirectory
            parentPath.endsWith(THEME_GROUPHOME_DIRECTORY) -> loadedGroupHomeImageDirectory
            else -> null
        }
        return destinationParent?.let { File(it, sourceFile.name) }
    }

    companion object {
        private val JSON_FILE_SUFFIX = ".json"
        @JvmStatic
        protected val THEME_IMAGE_DIRECTORY = "images"
        @JvmStatic
        protected val THEME_TIMELINE_DIRECTORY = "timeline"
        @JvmStatic
        protected val THEME_GROUPHOME_DIRECTORY = "groupboard"
        private val PATH_SEPARATOR = '/'

        @Throws(IOException::class)
        private fun extract(dstFile: File?, zipInputStream: ZipInputStream) {
            if (dstFile == null) {
                return
            }
            if (dstFile.exists()) {
                dstFile.delete()
            }

            try {
                FileOutputStream(dstFile).use { zipInputStream.copyTo(it) }
            } catch (e: IOException) {
                ThemeErrorReportUtil.report(
                        "Fail to copy file $dstFile",
                        "ThemeFileManager.extract(File, ZipINputStream)", e)
                dstFile.delete()
                throw e
            }
        }

        private fun deleteFilesRecursively(target: File?) {
            if (target == null || !target.exists()) {
                return
            }

            // In case of directory
            target.listFiles()?.forEach { file -> deleteFilesRecursively(file) }

            if (!target.delete()) {
                target.deleteOnExit()
            }
        }

        private fun SafeZipInputStream.forEachNonDirectoryFile(block: (ZipEntry) -> Unit) {
            var entry: ZipEntry? = nextEntry
            while (entry != null) {
                if (!entry.isDirectory) {
                    block(entry)
                }
                entry = nextEntry
            }
        }
    }
}
