@file:Suppress("DEPRECATION") // For SharedPrefWrapper import

package jp.naver.line.androie.common.theme

import android.graphics.PorterDuff
import android.graphics.drawable.Drawable
import android.support.annotation.ColorInt
import android.support.annotation.IdRes
import android.support.v4.graphics.drawable.DrawableCompat
import android.view.View
import jp.naver.line.androie.common.access.LineAccessForCommonHelper
import jp.naver.line.androie.common.theme.ThemeElementValue.ResourceType
import jp.naver.line.androie.sharedpref.SharedPrefKey
import jp.naver.line.androie.sharedpref.SharedPrefUtils
import jp.naver.line.androie.sharedpref.SharedPrefWrapper
import org.json.JSONException
import java.io.File
import java.io.IOException
import java.util.EnumMap
import java.util.HashMap

class ThemeManager @JvmOverloads constructor(
        val fileManager: ThemeFileManager,
        /**
         * True when this needs to apply the default theme before applying this theme.
         * This should be enabled for a theme which is the subset of a normal theme.
         */
        val isDefaultThemeRequired: Boolean = false
) {
    @Suppress("DEPRECATION")
    private val sharedPrefWrapper: SharedPrefWrapper? =
            SharedPrefUtils.open(SharedPrefKey.THEME_MANAGER)
    var currentPackageName: String = DEFAULT_THEME_PRODUCT_ID
        private set

    private var viewMapping: ThemeMappingData = ThemeMappingData.DEFAULT
    private var themeResources: ThemeResources = ThemeResources.DEFAULT
    private var themeElementMapping: MutableMap<ThemeElementKey, ThemeElementValue> = mutableMapOf()

    constructor(
            originalThemeManager: ThemeManager,
            vararg themeElements: Pair<Array<ThemeElementKey>, ThemeElementValue>
    ) : this(originalThemeManager.fileManager, originalThemeManager.isDefaultThemeRequired) {
        viewMapping = originalThemeManager.viewMapping
        currentPackageName = originalThemeManager.currentPackageName
        themeResources = ThemeResources(originalThemeManager.fileManager)
        themeElementMapping = HashMap(originalThemeManager.themeElementMapping)

        for ((themeElementKeys, themeElementValue) in themeElements) {
            for (themeElementKey in themeElementKeys) {
                themeElementMapping.put(themeElementKey, themeElementValue)
            }
        }
    }

    val lastValidationTime: Long
        get() = sharedPrefWrapper?.getLong(THEME_LAST_VALIDATION_TIME, 0L) ?: 0L

    fun updateLastValidationTime(time: Long) =
            sharedPrefWrapper?.putValue(THEME_LAST_VALIDATION_TIME, time)

    @Throws(IOException::class, JSONException::class)
    fun saveAndExtract(productId: String, zipFile: File?) {
        fileManager.flushThemeLoadFolder()

        val targetProductId =
                if (productId.isEmpty() || zipFile == null) DEFAULT_THEME_PRODUCT_ID else productId
        if (DEFAULT_THEME_PRODUCT_ID != targetProductId && zipFile != null) {
            fileManager.extract(zipFile)
        }
        load(targetProductId)
        sharedPrefWrapper?.putValue(THEME_PACKAGE_NAME, targetProductId)
    }

    /**
     * Returns true if the current theme is the default one.
     */
    val isDefaultTheme: Boolean
        get() = DEFAULT_THEME_PRODUCT_ID == currentPackageName

    /**
     * Loads a theme which is determined by shared preference. With calling this, the already loaded
     * theme will be overridden.
     */
    fun load() {
        // TODO: Refactor SharedPrefWrapper.
        val packageName = sharedPrefWrapper?.getString(THEME_PACKAGE_NAME, DEFAULT_THEME_PRODUCT_ID)
        // sharedPrefWrapper.getString() can return null if the null is set intentionally.
        try {
            load(packageName)
        } catch (e: IOException) {
            ThemeErrorReportUtil.report(
                    "Error while loading $packageName", "ThemeManager.load()", e)
        } catch (e: JSONException) {
            ThemeErrorReportUtil.report(
                    "Error while loading $packageName", "ThemeManager.load()", e)
        }
    }

    @Throws(IOException::class, JSONException::class)
    fun load(packageName: String?) {
        currentPackageName = packageName ?: DEFAULT_THEME_PRODUCT_ID

        val lineApi = LineAccessForCommonHelper.getLineAccess()
        viewMapping = ThemeMappingData(
                ThemeKey.values().associateNonNullValue(lineApi::getThemeMappingData))

        if (isDefaultTheme) {
            themeResources = ThemeResources.DEFAULT
            themeElementMapping = mutableMapOf()
        } else {
            themeResources = ThemeResources(fileManager)
            val loadedData = ThemeJsonLoader.load(fileManager.jsonFile)
            themeElementMapping = loadedData ?: mutableMapOf()
        }
    }

    // TODO: Remove this.
    @Deprecated("Use {@link #applyThemeMap(View, ThemeElementMappingData...)} instead.")
    fun applyTheme(rootView: View, vararg themeKeys: ThemeKey): Boolean {
        var applied = false
        for (themeKey in themeKeys) {
            @Suppress("DEPRECATION")
            applied = applied or applySingleThemeKey(rootView, themeKey)
        }
        return applied
    }

    // TODO: Remove this.
    @Deprecated("Use {@link #applyThemeMap(View, ThemeElementMappingData...)} instead.")
    fun applyTheme(rootView: View?, themeKey: ThemeKey): Boolean =
            applySingleThemeKey(rootView, themeKey)

    // TODO: Remove this.
    @Deprecated("Use {@link #applyTheme(View, ThemeElementKey, ThemeElementDefaultValue)} instead.")
    fun applyTheme(view: View, themeKey: ThemeKey, viewId: Int): Boolean {
        val elementMappingData = viewMapping.getMappingData(themeKey, viewId) ?: return false
        return applyTheme(
                view,
                ThemeElementValueMerger.getMergedValue(
                        themeElementMapping, *elementMappingData.elementKeys),
                elementMappingData.defaultValue,
                elementMappingData.mappingTargets)
    }

    private fun applySingleThemeKey(rootView: View?, themeKey: ThemeKey): Boolean {
        if (rootView == null) {
            return false
        }

        // TODO: refactor viewIds and itemInfos of ThemeMetaInfo. Because there is null info so that
        // the corresponding id is non-null, we can't loop with itemInfos.
        var isAppliedAny = false
        for (elementMappingData in viewMapping.iterableOf(themeKey)) {
            isAppliedAny = isAppliedAny or applyTheme(
                    rootView.findViewById(elementMappingData.resourceId),
                    ThemeElementValueMerger.getMergedValue(
                            themeElementMapping, *elementMappingData.elementKeys),
                    elementMappingData.defaultValue,
                    elementMappingData.mappingTargets)
        }
        return isAppliedAny
    }

    fun applyTheme(
            view: View,
            themeElementKey: ThemeElementKey,
            defaultValue: ThemeElementDefaultValue?
    ): Boolean = applyTheme(
            view,
            themeElementMapping[themeElementKey],
            defaultValue,
            ThemeElementMappingData.ALL_TARGETS)

    fun applyTheme(
            view: View,
            themeElementKeys: Array<ThemeElementKey>,
            defaultValue: ThemeElementDefaultValue? = null
    ): Boolean = applyTheme(
            view,
            ThemeElementValueMerger.getMergedValue(themeElementMapping, *themeElementKeys),
            defaultValue,
            ThemeElementMappingData.ALL_TARGETS)

    fun applyThemeMap(
            view: View,
            vararg elementMappingDataArray: ThemeElementMappingData): Boolean {
        var isAppliedAny = false
        for (elementMappingData in elementMappingDataArray) {
            isAppliedAny = isAppliedAny or applyTheme(
                    view.findViewById(elementMappingData.resourceId),
                    ThemeElementValueMerger.getMergedValue(
                            themeElementMapping, *elementMappingData.elementKeys),
                    elementMappingData.defaultValue,
                    elementMappingData.mappingTargets)
        }
        return isAppliedAny
    }

    fun getTintedImageDrawable(
            themeKey: Array<ThemeElementKey>,
            defaultDrawable: Drawable
    ): Drawable {
        val themeElementValue = getThemeElementValue(*themeKey)
        val drawable = themeElementValue.image?.getDrawable(themeResources) ?: defaultDrawable
        val wrappedDrawable = DrawableCompat.wrap(drawable)
        val colorStateList = themeElementValue.imageTintColor?.colorStateList
        DrawableCompat.setTintList(wrappedDrawable, colorStateList)
        return wrappedDrawable
    }

    private fun applyTheme(
            view: View?,
            themeValue: ThemeElementValue?,
            defaultValue: ThemeElementDefaultValue?,
            applyingTargets: Collection<ThemeElementValueType>): Boolean {
        if (view == null) {
            return false
        }

        if (isDefaultThemeRequired) {
            defaultValue?.applyDefaultTheme(view)
        }
        val isApplied = themeValue != null &&
                themeValue.applyTheme(view, themeResources, applyingTargets)
        if (!isDefaultThemeRequired && !isApplied) {
            defaultValue?.applyDefaultTheme(view)
        }
        return isApplied
    }

    /**
     * Applies a tint color specified in the given theme to the Drawable.
     *
     * @param key a Theme Key
     * @param resourceId a resource id to obtain the tint color from the Theme
     * @param drawable a Drawable to be applied the tint color
     * @return a Drawable that has been applied the tint color
     */
    // TODO : Remove this.
    @Deprecated("Use {@link #applyImageTintColor(Drawable, ThemeElementKey)} instead.")
    fun applyImageTintColor(key: ThemeKey?, resourceId: Int, drawable: Drawable): Drawable {
        @Suppress("DEPRECATION")
        val imageTintColor = getThemeElementValue(key, resourceId).imageTintColor ?: return drawable

        val drawableCompat = DrawableCompat.wrap(drawable.mutate())
        DrawableCompat.setTintMode(drawableCompat, PorterDuff.Mode.SRC_IN)
        DrawableCompat.setTint(drawableCompat, imageTintColor.colorInt)
        return drawableCompat
    }

    // TODO: Remove this.
    @Deprecated("Use {@link #getThemeElementValue(ThemeElementKey)} instead.")
    fun getDrawable(themeKey: ThemeKey, resId: Int): Drawable? {
        @Suppress("DEPRECATION")
        val themeValue = getThemeElementValue(themeKey, resId)
        val imageInfo = themeValue.image
        return if (imageInfo != null) {
            getDrawable(imageInfo)
        } else {
            themeValue.backgroundImage?.let { getDrawable(it) }
        }
    }

    // TODO: Remove this.
    @Deprecated("Use {@link #getThemeElementValue(ThemeElementKey)} instead.")
    fun getColor(themeKey: ThemeKey?, @IdRes viewId: Int): Map<ResourceType, Int>? {
        @Suppress("DEPRECATION")
        val itemInfo = getThemeElementValue(themeKey, viewId)
        if (itemInfo.isEmpty) {
            return null
        }

        val result = EnumMap<ResourceType, Int>(ResourceType::class.java)
        toDefaultColor(itemInfo.backgroundColor)
                ?.let { result.put(ResourceType.BG_COLOR, it) }
        toDefaultColor(itemInfo.textColor)
                ?.let { result.put(ResourceType.TEXT_COLOR, it) }
        return result
    }

    // TODO: Remove this.
    @Deprecated("Use {@link #getThemeElementValue(ThemeElementKey)} instead.")
    fun getThemeElementValue(themeKey: ThemeKey?, @IdRes viewId: Int): ThemeElementValue {
        val elementMappingData = themeKey?.let { viewMapping.getMappingData(it, viewId) }
                ?: return ThemeElementValue.INVALID
        val themeElementValue = ThemeElementValueMerger.getMergedValue(
                themeElementMapping, *elementMappingData.elementKeys)
        return themeElementValue ?: ThemeElementValue.INVALID
    }

    fun getThemeElementValue(vararg themeElementKeys: ThemeElementKey): ThemeElementValue =
            ThemeElementValueMerger.getMergedValue(themeElementMapping, *themeElementKeys)
                    ?: ThemeElementValue.INVALID

    private fun getDrawable(imageInfo: ThemeElementImageData?): Drawable? {
        if (imageInfo == null) {
            return null
        }

        var retryCount = 0
        while (retryCount <= 1) {
            try {
                return themeResources.getDrawable(imageInfo)
            } catch (e: Exception) {
                // When OutOfMemory or some other error happens, try to reload after clearing the
                // cache.
                // TODO: Use more specific Error.
                ThemeErrorReportUtil.report("Error while getting drawable $imageInfo",
                        "ThemeManager.getDrawable(ImageInfo)", e)
                themeResources.clearCache()
                retryCount++
            }
        }
        return null
    }

    companion object {
        @JvmField
        val DEFAULT_THEME_PRODUCT_ID = "3e261192-3a69-4849-b35d-35aeddd5a368"

        private val THEME_PACKAGE_NAME = "ThemePackageName"
        private val THEME_LAST_VALIDATION_TIME = "TLVT"

        /**
         * The default instance managing a theme which is determined by the shared preferences.
         */
        // TODO: Rename. Should not use singleton.
        @JvmStatic
        val instance: ThemeManager by lazy { ThemeManager(ApplicationThemeFileManager()) }

        @ColorInt
        private fun toDefaultColor(colorInfo: ThemeElementColorData?): Int? = colorInfo?.colorInt

        private inline fun <K, V> Array<K>.associateNonNullValue(
                transform: (K) -> V?
        ): Map<K, V> = mutableMapOf<K, V>().also { destination ->
            for (key in this) {
                transform(key)?.let { value -> destination[key] = value }
            }
        }
    }
}
