package jp.naver.line.androie.common.theme

import android.util.SparseArray
import java.util.EnumMap
import java.util.NoSuchElementException

/**
 * Data class to represent all mapping data of resource ids to theme elements.
 */
class ThemeMappingData(
        private val mappingData: Map<ThemeKey, SparseArray<ThemeElementMappingData>>) {

    private class SparseArrayIterable(
            private val sparseArray: SparseArray<ThemeElementMappingData>?
    ) : Iterable<ThemeElementMappingData> {
        override fun iterator(): Iterator<ThemeElementMappingData> =
                sparseArray?.let { SparseArrayIterator(sparseArray) } ?: EMPTY_MAPPING_DATA_ITERATOR
    }

    private class SparseArrayIterator<T>(private val sparseArray: SparseArray<T>) : Iterator<T> {
        private var currentIndex = 0

        override fun hasNext(): Boolean = sparseArray.size() > currentIndex

        override fun next(): T =
                if (hasNext()) {
                    sparseArray.valueAt(currentIndex++)
                } else {
                    throw NoSuchElementException(
                            "This iterator has no more element. index=$currentIndex")
                }
    }

    fun getMappingData(themeKey: ThemeKey, resourceId: Int): ThemeElementMappingData? =
            mappingData[themeKey]?.get(resourceId)

    fun iterableOf(themeKey: ThemeKey): Iterable<ThemeElementMappingData> =
            SparseArrayIterable(mappingData[themeKey])

    class Builder {
        private val mappingData: MutableMap<ThemeKey, SparseArray<ThemeElementMappingData>> =
                EnumMap(ThemeKey::class.java)

        fun addComponentMappingData(
                themeKey: ThemeKey,
                resourceIdToMappingData: SparseArray<ThemeElementMappingData>): Builder = apply {
            mappingData[themeKey] = resourceIdToMappingData
        }

        fun build(): ThemeMappingData = ThemeMappingData(mappingData.toMap())
    }

    class EmptyIterator<T> : Iterator<T> {
        override fun hasNext(): Boolean = false

        override fun next(): T =
                throw NoSuchElementException("next() is called for empty iterator")
    }

    companion object {
        val DEFAULT = ThemeMappingData(emptyMap())

        val EMPTY_MAPPING_DATA_ITERATOR: Iterator<ThemeElementMappingData> = EmptyIterator()
    }
}
