@file:JvmName("DisplayUtils")

package jp.naver.line.androie.common.util

/**
 * A collection of system display related utility methods.
 *
 * @author Hwan Jo(hwan.jo@linecorp.com)
 */

import android.app.Activity
import android.app.Application
import android.app.Dialog
import android.content.Context
import android.graphics.Rect
import android.os.Build
import android.support.annotation.ColorInt
import android.support.v4.content.ContextCompat
import android.util.DisplayMetrics
import android.view.View
import android.view.Window
import android.view.WindowManager
import android.view.inputmethod.InputMethodManager
import jp.naver.android.common.R
import jp.naver.line.androie.common.theme.CommonThemeKeys
import jp.naver.line.androie.common.theme.ThemeManager
import org.jetbrains.anko.displayMetrics
import org.jetbrains.anko.inputMethodManager
import kotlin.math.roundToInt

@Suppress("DEPRECATION")
private typealias DeprecatedApplicationKeeper = jp.naver.line.androie.common.ApplicationKeeper

@Suppress("TYPEALIAS_EXPANSION_DEPRECATION")
private val deprecatedApplication: Application
    get() = DeprecatedApplicationKeeper.getApplication()

/**
 * Calculates and returns the height of status bar.
 *
 * @return height of system status bar in pixels on currently running device.
 */
@Deprecated(
        "Use an extension \"Context.statusBarHeightInPixel\".",
        replaceWith = ReplaceWith("val Context.statusBarHeightInPixel: Int")
)
fun getStatusBarHeight(): Int = deprecatedApplication.statusBarHeightInPixel

/**
 * @return display density of currently running device.
 */
@Deprecated(
        "Use an extension \"Context.displayDensity\".",
        replaceWith = ReplaceWith("val Context.displayDensity: Float")
)
fun getDisplayDensity(): Float = deprecatedApplication.displayDensity

/**
 * @return The screen density expressed as dots-per-inch.
 */
@Deprecated(
        "Use an extension \"Context.displayDensityDpi\".",
        replaceWith = ReplaceWith("val Context.displayDensityDpi: Int")
)
fun getDisplayDensityDpi(): Int = deprecatedApplication.displayDensityDpi

/**
 * @return the absolute width of the display in pixels.
 */
@Deprecated(
        "Use an extension \"Context.displayWidthInPixel\".",
        replaceWith = ReplaceWith("val Context.displayWidthInPixel: Int")
)
fun getDisplayWidth(): Int = deprecatedApplication.displayWidthInPixel

/**
 * @return the absolute height of the display in pixels.
 */
@Deprecated(
        "Use an extension \"Context.displayHeightInPixel\".",
        replaceWith = ReplaceWith("val Context.displayHeightInPixel: Int")
)
fun getDisplayHeight(): Int = deprecatedApplication.displayHeightInPixel

/**
 * @return pixel size of given `dip` on currently running device.
 */
@Deprecated(
        "Use an overload \"toPixel(context, dip)\".",
        replaceWith = ReplaceWith("fun toPixel(context: Context, dip: Float): Int")
)
fun toPixel(dip: Float): Int = toPixel(deprecatedApplication, dip)

/**
 * @return pixel size of given `dip` on currently running device.
 * For Kotlin codes, It is strongly recommended to use [org.jetbrains.anko.dip] extensions
 * declared by Jetbrain instead of [toPixel].
 *
 * @see org.jetbrains.anko.dip
 */
fun toPixel(context: Context, dip: Float): Int = (dip * context.displayDensity).roundToInt()

/**
 * @return converts given `pixel` to dip unit on currently running device.
 */
@Deprecated(
        "Use an overload \"toDip(context, pixel)\".",
        replaceWith = ReplaceWith("fun toDip(context: Context, pixel: Int): Int")
)
fun toDip(pixel: Int): Float = toDip(deprecatedApplication, pixel)

/**
 * @return converts given `pixel` to dip unit on currently running device.
 * For Kotlin codes, It is strongly recommended to use [org.jetbrains.anko.px2dip] extensions
 * declared by Jetbrain instead of [toDip].
 *
 * @see org.jetbrains.anko.px2dip
 */
fun toDip(context: Context, pixel: Int): Float = (pixel / context.displayDensity)

/**
 * Return the screen display metrics, containing absolute width/height in pixels.
 *
 * For Kotlin codes, It is strongly recommended to use [org.jetbrains.anko.displayMetrics] extensions
 * declared by Jetbrain instead of [getDisplayMetrics].
 */
fun getDisplayMetrics(context: Context): DisplayMetrics {
    var dm = context.resources.displayMetrics
    if (dm.widthPixels <= 0 || dm.heightPixels <= 0) {
        // Workaround for ICS https://stackoverflow.com/a/11004877/5489877
        val wm = context.getSystemService(Context.WINDOW_SERVICE) as WindowManager
        dm = DisplayMetrics()
        wm.defaultDisplay.getMetrics(dm)
    }
    return dm
}

/**
 * Forces the input window to be displayed whether there are input controls or not on current receiver.
 *
 * This will do nothing if the given receiver is ``null``.
 */
@JvmOverloads
fun Context?.showSoftInput(optionalFlags: Int = 0) =
        this?.inputMethodManager?.toggleSoftInput(optionalFlags, InputMethodManager.SHOW_IMPLICIT)
                ?: Unit

/**
 * Forces the input window to be hidden on current `activity`
 *
 * This method will do nothing if either `activity` or `view` is/are `null`.
 */
@JvmOverloads
fun Context?.hideSoftInput(view: View?, optionalFlags: Int = 0): Boolean {
    val windowToken = view?.windowToken ?: return false
    return this?.inputMethodManager?.hideSoftInputFromWindow(windowToken, optionalFlags) == true
}

/**
 * Height of system status bar in pixels on currently running device.
 */
val Context.statusBarHeightInPixel: Int
    get() = if (this is Activity) {
        Rect().also(window.decorView::getWindowVisibleDisplayFrame).top
    } else {
        resources
                .getIdentifier("status_bar_height", "dimen", "android")
                .takeIf { it > 0 }
                ?.let(resources::getDimensionPixelSize)
                ?: 0
    }

/**
 * Display density of currently running device.
 */
val Context.displayDensity: Float get() = displayMetrics.density

/**
 * The screen density expressed as dots-per-inch.
 *
 * @see DisplayMetrics
 */
val Context.displayDensityDpi: Int get() = displayMetrics.densityDpi

/**
 * The absolute width of the display in pixels.
 */
val Context.displayWidthInPixel: Int get() = displayMetrics.widthPixels

/**
 * The absolute height of the display in pixels.
 */
val Context.displayHeightInPixel: Int get() = displayMetrics.heightPixels

fun Window.changeStatusBarColor(@ColorInt newStatusBarColor: Int) {
    if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
        if (!isFlagEnabled(WindowManager.LayoutParams.FLAG_DRAWS_SYSTEM_BAR_BACKGROUNDS)) {
            // change status bar color
            addFlags(WindowManager.LayoutParams.FLAG_DRAWS_SYSTEM_BAR_BACKGROUNDS)
        }
        statusBarColor = newStatusBarColor
    }
}

fun Activity.changeStatusBarColor(@ColorInt newStatusBarColor: Int) =
        window.changeStatusBarColor(newStatusBarColor)

@JvmOverloads
fun Activity.changeStatusBarColorWithTheme(theme: ThemeManager = ThemeManager.instance) =
        window.changeStatusBarColor(loadStatusBarColorFrom(theme))

@JvmOverloads
fun Dialog.changeStatusBarColorWithTheme(theme: ThemeManager = ThemeManager.instance) =
        window.changeStatusBarColor(context.loadStatusBarColorFrom(theme))

@JvmOverloads
fun View.changeStatusBarColorWithTheme(theme: ThemeManager = ThemeManager.instance) =
        (context as? Activity)?.changeStatusBarColorWithTheme(theme) ?: Unit

@ColorInt
private fun Context.loadStatusBarColorFrom(theme: ThemeManager): Int {
    // Get status bar color from theme.
    val navigationBarColor = theme
            .getThemeElementValue(*CommonThemeKeys.StatusBar.ROOT)
            .backgroundColor
    // Fallback to default_status_bar_bg if theme color is not available.
    return navigationBarColor?.colorInt
            ?: ContextCompat.getColor(this, R.color.default_status_bar_bg)
}

private fun Window.isFlagEnabled(flags: Int): Boolean = (attributes.flags and flags) == flags
