package jp.naver.line.androie.common.view

import android.content.Context
import android.content.res.ColorStateList
import android.graphics.Canvas
import android.graphics.PorterDuff
import android.graphics.drawable.Drawable
import android.support.annotation.ColorInt
import android.support.annotation.StyleableRes
import android.support.v4.graphics.drawable.DrawableCompat
import android.support.v4.widget.ImageViewCompat
import android.support.v7.widget.AppCompatImageView
import android.util.AttributeSet
import android.widget.ImageView
import jp.naver.android.common.R
import jp.naver.line.androie.util.runWithTypedArray

/**
 * A tintable [ImageView], which has an overwrapping drawable. If a drawable is given by
 * [setOverwrappingDrawable] or an xml attribute `overwrappingSrc`, the
 * drawable is shown when this view is tinted.
 *
 *
 * When an image is tinted with [PorterDuff.Mode.SRC_ATOP] and without alpha, the image is
 * just filled with a color. For example, a check mark icon "`[v]`" will look like "`[ ]`" after
 * tinting. This view can stack another drawable layer after tinting. You can crate an image with
 * two colors even with tinting. For example, when a check mark drawable "` v `" is provided in
 * addition to an icon "`[v]`", this overwraps the tinted icon "`[ ]`" with the given mark "` v `".
 * Then, an overwrapped tinted icon "`[v]`" is shown.
 *
 *
 * If xml attribute `drawOverwrappingImageWithoutTint` is specified, the overwrapping
 * drawable will be shown even if this is not tinted.
 */
// TODO: Rename interface ImageTintableView to TintableImage.
class OverwrappedTintableImageView @JvmOverloads constructor(
        context: Context,
        attrs: AttributeSet? = null,
        defStyleAttr: Int = 0
) : AppCompatImageView(context, attrs, defStyleAttr) {

    private val isOriginalDrawableTinted: Boolean
        get() = ImageViewCompat.getImageTintList(this) != null

    /**
     * Overwrapped drawable on the [ImageView] original drawable. This is rendered if the following
     * conditions are satisfied.
     * 1. [isOverwrappingDrawableEnabled] is true
     * 2. This view is tinted, or [shouldDrawOverwrappingImageWithoutTint] is true.
     */
    private var overwrappingDrawable: Drawable? = null

    /**
     * Tint color for [overwrappingDrawable]. Though this type is [ColorStateList], only the normal
     * (= default) state is used.
     */
    private var overwrappingDrawableTintColorList: ColorStateList? = null

    private var shouldDrawOverwrappingImageWithoutTint: Boolean = false
    private var isOverwrappingDrawableEnabled: Boolean = true

    init {
        attrs?.runWithTypedArray(
                context, R.styleable.OverwrappedTintableImageView, defStyleAttr) { typedArray ->
            overwrappingDrawable =
                    typedArray.getDrawable(OVERWRAPPING_SRC_STYLEABLE_RES)?.toTintable()
            overwrappingDrawableTintColorList =
                    typedArray.getColorStateList(OVERWRAPPING_TINT_STYLEABLE_RES)
            shouldDrawOverwrappingImageWithoutTint = typedArray
                    .getBoolean(SHOW_OVERWRAPPING_IMAGE_WITHOUT_TINT_STYLEABLE_RES, false)
        }
    }

    /**
     * Sets an overwrapping drawable, which is shown when this image view is tinted.
     */
    fun setOverwrappingDrawable(overwrappingDrawable: Drawable?) {
        this.overwrappingDrawable?.let { drawable ->
            drawable.callback = null
            unscheduleDrawable(drawable)
        }
        this.overwrappingDrawable = overwrappingDrawable?.toTintable()
        updateOverwrappingDrawableTint()
    }

    /**
     * Sets a tint color for an overwrapping drawable, which is set by
     * [setOverwrappingDrawable].
     */
    fun setOverwrappingDrawableTintColor(@ColorInt tintColor: Int) {
        overwrappingDrawableTintColorList = ColorStateList.valueOf(tintColor)
        updateOverwrappingDrawableTint()
    }

    /**
     * Clears a tint color for an overwrapping drawable where the tint color is given by
     * [setOverwrappingDrawableTintColor] or a xml attribute.
     */
    fun clearOverwrappingDrawableTintColor() {
        overwrappingDrawableTintColorList = null
        updateOverwrappingDrawableTint()
    }

    fun setOverwrappingDrawableEnabled(isOverwrappingDrawableEnabled: Boolean) {
        this.isOverwrappingDrawableEnabled = isOverwrappingDrawableEnabled
    }

    override fun draw(canvas: Canvas) {
        super.draw(canvas)
        val drawable = overwrappingDrawable ?: return
        if (isOverwrappingDrawableEnabled &&
                (isOriginalDrawableTinted || shouldDrawOverwrappingImageWithoutTint)) {
            drawable.draw(canvas)
        }
    }

    override fun onSizeChanged(w: Int, h: Int, oldw: Int, oldh: Int) {
        super.onSizeChanged(w, h, oldw, oldh)
        overwrappingDrawable?.let {
            it.setBounds(paddingLeft, paddingTop, w - paddingRight, h - paddingBottom)
            invalidate()
        }
    }

    override fun verifyDrawable(drawable: Drawable): Boolean =
            overwrappingDrawable === drawable || super.verifyDrawable(drawable)

    private fun updateOverwrappingDrawableTint() = overwrappingDrawable?.let { drawable ->
        DrawableCompat.setTintList(drawable, overwrappingDrawableTintColorList)
        invalidate()
    }

    private fun Drawable.toTintable(): Drawable = DrawableCompat.wrap(this).mutate()

    companion object {
        @StyleableRes
        private val OVERWRAPPING_SRC_STYLEABLE_RES =
                R.styleable.OverwrappedTintableImageView_overwrappingSrc
        @StyleableRes
        private val OVERWRAPPING_TINT_STYLEABLE_RES =
                R.styleable.OverwrappedTintableImageView_overwrappingTint
        @StyleableRes
        private val SHOW_OVERWRAPPING_IMAGE_WITHOUT_TINT_STYLEABLE_RES =
                R.styleable.OverwrappedTintableImageView_drawOverwrappingImageWithoutTint
    }
}
