package jp.naver.line.androie.common.view

import android.content.Context
import android.content.res.ColorStateList
import android.graphics.Color
import android.graphics.PorterDuff
import android.graphics.drawable.Drawable
import android.support.annotation.ColorInt
import android.util.AttributeSet
import android.widget.ProgressBar
import jp.naver.android.common.R

/**
 * A [ProgressBar] which implements [ImageTintable].
 *
 * This view does not support [ColorStateList], only supports a single color tinting.
 */
class TintableProgressBar : ProgressBar, ImageTintable {

    @ColorInt
    private var color: Int

    // The default defStyleAttr for `constructor(Context, AttributeSet?)` is
    // `com.android.internal.R.attr.progressBarStyle` while `0` is used for usual views.
    // Then, we have to call `super(Context, AttributeSet?)` for this constructor because the
    // resource is hidden.
    // Thus, we cannot use default value to construct this object.
    @JvmOverloads constructor(
            context: Context,
            attrs: AttributeSet? = null
    ) : super(context, attrs) {
        color = colorIntFrom(attrs, 0)
        updateTintColor()
    }

    constructor(
            context: Context,
            attrs: AttributeSet?,
            defStyleAttr: Int
    ) : super(context, attrs, defStyleAttr) {
        color = colorIntFrom(attrs, defStyleAttr)
        updateTintColor()
    }

    override fun setIndeterminate(indeterminate: Boolean) {
        super.setIndeterminate(indeterminate)
        updateTintColor()
    }

    override fun setProgressDrawable(d: Drawable?) {
        super.setProgressDrawable(d)
        updateTintColor()
    }

    override fun setIndeterminateDrawable(d: Drawable?) {
        super.setIndeterminateDrawable(d)
        updateTintColor()
    }

    override fun tintImage(colors: ColorStateList) {
        color = colors.defaultColor
        updateTintColor()
    }

    @ColorInt
    private fun colorIntFrom(attrs: AttributeSet?, defStyleAttr: Int): Int {
        if (attrs == null) {
            return Color.TRANSPARENT
        }
        val typedArray = context.theme.obtainStyledAttributes(
                attrs, R.styleable.TintableProgressBar, defStyleAttr, 0)
        val colorInt =
                typedArray.getColor(R.styleable.TintableProgressBar_tintColor, Color.TRANSPARENT)
        typedArray.recycle()
        return colorInt
    }

    private fun updateTintColor() {
        indeterminateDrawable?.setColorFilter(color, PorterDuff.Mode.SRC_ATOP)
        progressDrawable?.setColorFilter(color, PorterDuff.Mode.SRC_ATOP)
    }
}
