package jp.naver.line.androie.util

import android.content.Context
import android.content.res.Resources
import android.os.Build
import android.support.v4.util.LruCache
import android.text.format.DateUtils
import android.util.Log
import java.util.Locale

/**
 * Utility Class to get relative time span string such as "Today", "Yesterday", "3 days ago", etc...
 *
 * This is a workaround for android's standard DateUtils.getRelativeTimeSpanString()
 * which has a bug on android L or previous:
 * It always returns strings in device's language instead of context's language.
 *
 * Warning: This class is not thread-safe.
 */
class DateUtilsWorkaroundCachedStringDataManager {

    /**
     * Map to hold strings in memory so that we can reuse them later
     * to minimize the number of the risky hacks as much as possible.
     *
     * Keys: Instances of CacheKey
     * Values: Relative time span strings such as "Today", "Yesterday", "3 days ago", etc...
     */
    private val cachedLocalizedRelativeDayStrings: LruCache<CacheKey, CharSequence> by lazy {
        LruCache<CacheKey, CharSequence>(CACHE_SIZE)
    }

    fun clearAllCachedStrings() {
        cachedLocalizedRelativeDayStrings.evictAll()
    }

    /**
     *  Get relative day string such as "Today", "Yesterday", "3 days ago", etc...
     */
    fun getRelativeDayString(
            context: Context,
            targetTimeMillis: Long,
            currentTimeMillis: Long,
            flags: Int
    ): CharSequence {
        if (!shouldUseCache(context)) {
            return DateUtils.getRelativeTimeSpanString(
                    targetTimeMillis, currentTimeMillis, DateUtils.DAY_IN_MILLIS, flags)
        }

        val key = CacheKey(localeFromResources(context.resources),
                flags, targetTimeMillis, currentTimeMillis)

        var relativeTimeString = cachedLocalizedRelativeDayStrings.get(key)

        // If cached relative time string exists, returns it
        if (relativeTimeString != null) {
            return relativeTimeString
        }

        // Otherwise, creates a time string and caches it.
        val backupDeviceLocale = localeFromResources(Resources.getSystem())
        setDeviceLocale(localeFromResources(context.resources))
        relativeTimeString = DateUtils.getRelativeTimeSpanString(
                targetTimeMillis, currentTimeMillis, DateUtils.DAY_IN_MILLIS, flags)
        setDeviceLocale(backupDeviceLocale)
        cachedLocalizedRelativeDayStrings.put(key, relativeTimeString)

        Log.d(TAG, "getRelativeDayString: key = '$key'; result = '$relativeTimeString'")

        return relativeTimeString
    }

    /**
     * Key to use with cachedLocalizedRelativeDayStrings:
     *      - localeString: locale string such as ja, en, vi, es, etc...
     *      - flags: Format flags: is one or multiple combined flags defined in android's DateUtils.
     *      - days: Int value such as: 1 day, 2 days, etc...
     */
    private data class CacheKey(
            val localeString: String,
            val flags: Int,
            val days: Int
    ) {
        constructor(
                appLocale: Locale,
                flags: Int,
                targetTimeMillis: Long,
                currentTimeMillis: Long
        ) : this(
                appLocale.toString(),
                flags,
                DateFormatUtil.getDaysBetween(targetTimeMillis, currentTimeMillis).toInt()
        )
    }

    companion object {
        private const val TAG: String = "DateUtilsLocalizeFix"

        private const val CACHE_SIZE: Int = 20

        private fun setDeviceLocale(targetLocale: Locale) =
                if (Build.VERSION.SDK_INT < Build.VERSION_CODES.N) {
                    @Suppress("DEPRECATION") // Suppress to avoid Kotlin checker warning
                    Resources.getSystem().configuration.locale = targetLocale
                } else {
                    Resources.getSystem().configuration.setLocale(targetLocale)
                }

        private fun localeFromResources(resources: Resources): Locale =
                if (Build.VERSION.SDK_INT < Build.VERSION_CODES.N) {
                    @Suppress("DEPRECATION") // Suppress to avoid Kotlin checker warning
                    resources.configuration.locale
                } else {
                    resources.configuration.locales[0]
                }

        private fun shouldUseCache(context: Context) =
                localeFromResources(context.resources).language !=
                        localeFromResources(Resources.getSystem()).language
    }
}
