package jp.naver.line.androie.util

/**
 * [Lazy] object which can be revoked. Unlike with a normal lazy object created by [lazy],
 * [value] is re-initialized when [revoke] is called.
 *
 * Usage:
 * ```kotlin
 * var count = 0
 * val lazyObject = revocableLazy { ++count }
 *
 * lazyObject.value //=> 1
 * lazyObject.value //=> 1
 *
 * lazyObject.revoke()
 *
 * lazyObject.value //=> 2
 * lazyObject.value //=> 2
 * ```
 *
 * Caution: Unlike with a normal lazy object, this lazy has the following conditions.
 * 1. [value] might differ before and after [revoke] call
 * 2. [initializer] might be called multiple times
 * 3. [initializer] won't be released after initialization.
 *
 * Note: This class is thread unsafe.
 */
class RevocableLazy<out T>(private val initializer: () -> T) : Lazy<T> {

    // Note: Because `lateinit` cannot be used with nullable upper-bound, we need to use `Any`.
    private var shadowValue: Any? = InvalidValue

    /**
     * Revokes the current [value] if it's already initialized. After this method call, the given
     * [initializer] will be called again when [value] is accessed.
     */
    fun revoke() {
        shadowValue = InvalidValue
    }

    override fun isInitialized(): Boolean = shadowValue !== InvalidValue

    override val value: T
        get() {
            if (!isInitialized()) {
                shadowValue = initializer()
            }
            @Suppress("UNCHECKED_CAST")
            return shadowValue as T
        }

    companion object {
        private object InvalidValue
    }
}

fun <T> revocableLazy(initializer: () -> T): RevocableLazy<T> = RevocableLazy(initializer)
